package sim.application.util;

import java.awt.BorderLayout;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;

import javax.swing.JFrame;
import javax.swing.JPanel;
import javax.swing.border.EmptyBorder;

import sim.exception.SConstructorException;
import sim.geometry.space.SAbstractGeometrySpace;
import sim.graphics.SRaytracer;
import sim.graphics.SScene;
import sim.util.SChronometer;
import sim.util.SLog;
import javax.swing.JLabel;
import javax.swing.JTextField;
import javax.swing.JButton;
import javax.swing.JFileChooser;

import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import java.awt.GridLayout;
import javax.swing.SwingConstants;

/**
 * La classe <b>SJFrameRenderer</b> reprsentant un JFrame permettant de lancer l'application du <i>ray tracer</i>.
 * 
 * @author Simon Vzina
 * @since 2015-08-22
 * @version 2023-02-03
 */
public class SJFrameRenderer extends JFrame {

  /**
   * La variable <b>NB_LINES_BEFORE_REPAINT</b> reprsente le nombre de lignes calcules par le <i>ray tracer</i> avant d'effectuer un repaint du jpanel.
   */
  private final int NB_LINES_BEFORE_REPAINT = 1;   
  
  /**
   * 
   */
  private static final long serialVersionUID = 1663616456131809383L;
  
  /**
   * La variable <b>configuration</b> reprsente la configuration du rendu.
   */
  private SConfiguration configuration;
  
  private JPanel contentPane;
  private JLabel labelFileName;
  private JTextField textFieldFileName;
  private SJPanelImage panelImage;
  private JButton btnBrowser;

  //----------------
  // CONSTRUCTEUR //
  //----------------
  
  /**
   * ...
   * 
   * @param configuration
   */
  public SJFrameRenderer(SConfiguration configuration)
  {
    this();
    
    this.configuration = configuration;
    
    this.textFieldFileName.setText(this.configuration.getReadDataFileName());
  }
  
  /**
   * Constructeur de l'application <i>ray tracer</i> dans un format "JFrame".
   */
  private SJFrameRenderer()
  {
    setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
    
    setBounds(100, 100, 640, 480);
    
    //Le panel du renderer
    contentPane = new JPanel();
    contentPane.setBorder(new EmptyBorder(5, 5, 5, 5));
    contentPane.setLayout(new BorderLayout(0, 0));
    contentPane.setPreferredSize(getMaximumSize());       //permettre au jpanel d'occuper le maximum d'espace dans le jframe
    setContentPane(contentPane);
    
    panelImage = new SJPanelImage();
    contentPane.add(panelImage, BorderLayout.CENTER);
    
    JPanel panelSouth = new JPanel();
    contentPane.add(panelSouth, BorderLayout.SOUTH);
    panelSouth.setLayout(new GridLayout(0, 4, 0, 0));
    
    labelFileName = new JLabel("Nom du fichier : ");
    labelFileName.setHorizontalAlignment(SwingConstants.CENTER);
    panelSouth.add(labelFileName);
    
    textFieldFileName = new JTextField();
    panelSouth.add(textFieldFileName);
    textFieldFileName.setColumns(20);
    
    JButton buttonRaytrace = new JButton("Ray tracing");
    buttonRaytrace.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        
        // Mettre le raytracing dans un nouveau thread afin d'excuter le repaint
        // avant la fin de l'excution de l'action.
        new Thread() {
          public void run() {
            raytrace();
          }
      }.start();
      }
    });
    
    panelSouth.add(buttonRaytrace);
    
    btnBrowser = new JButton("Browser ...");
    btnBrowser.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        actionBrowser();
      }
    });
    panelSouth.add(btnBrowser);
  }

  //------------
  // MTHODES //
  //------------
   
  /**
   * Mthode pour effectuer le ray tracing de la scne
   */
  private void raytrace()
  {
    raytrace(this.textFieldFileName.getText(), this.configuration.getWriteDataFileName());
  }
  
  /**
   * Mthode pour effectuer le ray tracing de la scne.
   * 
   * @param file_name Le nom du fichier de la scne.
   * @param write_file_name Le nom du fichier en criture de la scne (version comprise dans l'engin graphique).
   */
  private void raytrace(String file_name, String write_file_name)
  {
    try
    {
      SChronometer chrono = new SChronometer();
      
      // Lecture de la scne
      chrono.start();
      SLog.logWriteLine("Message SJFrameRenderer : Lecture de la scne '" + file_name + "'.");
      
      // Dfinition de la scene.
      SScene scene = new SScene(file_name);
        
      chrono.stop();
      SLog.logWriteLine("Message SJFrameRenderer : Fin de la lecture de la scne.");
      SLog.logWriteLine("Message SJFrameRenderer : Dure : " + chrono.getTime() + " s.");
      SLog.logWriteLine("");
      
      //Ray tracing
      chrono.start();
      SLog.logWriteLine("Message SJFrameRenderer : Dbut du raytracing.");
      
      //Construction du raytracer
      SRaytracer raytracer = scene.buildRaytracer();
      
      // Mettre  jour la taille de la fentre
      this.setBounds(100, 100, scene.getViewport().getWidth(), scene.getViewport().getHeight());
            
      // Mettre  jour le titre de l'application.
      this.setTitle("Fichier : " + file_name);
      
      // Effectuer le ray tracing ligne par ligne
      int nb_pixels = scene.getViewport().getWidth()*NB_LINES_BEFORE_REPAINT;
      
      while(scene.getViewport().hasNextPixel())
      {
        raytracer.raytrace(nb_pixels);
        this.panelImage.setBufferedImage(scene.getViewport().getBufferedImage());
      }
                       
      chrono.stop();
      
      SLog.logWriteLine("Message SJFrameRenderer : Fin du raytracing.");
      SLog.logWriteLine("Message SJFrameRenderer : Dure : " + chrono.getTime() + " s.");
      SLog.logWriteLine("Message SJFrameRenderer : Nombre de tests d'intersection raliss : " + SAbstractGeometrySpace.getIntersectionTestCount() + " tests.");
      
      SLog.logWriteLine("");
      
      //criture de l'image
      chrono.start();
      SLog.logWriteLine("Message SJFrameRenderer : criture de l'image.");
      
      //criture du fichier png
      scene.getViewport().writeImage();
      
      chrono.stop();
      SLog.logWriteLine("Message SJFrameRenderer : Fin de l'criture de l'image.");
      SLog.logWriteLine("Message SJFrameRenderer : Dure : " + chrono.getTime() + " s.");
      SLog.logWriteLine("");
      
      // criture de la scne (interprtation crite de ce qui a t vritablement dessin).
      scene.write(write_file_name);
 
      // criture du fichier de configuration (sauvegarder le comptage des images).
      this.configuration.writeFile();    
      
    }catch(FileNotFoundException e){       
      SLog.logWriteLine("Message SJFrameRenderer : Une erreur de type FileNotFoundException est survenue.");
      SLog.logException(e);
    }catch(SConstructorException e){
      SLog.logWriteLine("Message SJFrameRenderer : Une erreur de type SConstructorException est survenue."); 
      SLog.logWriteLine(e.getMessage());
      e.printStackTrace();
    }catch(IOException e){                 
      SLog.logWriteLine("Message SJFrameRenderer : Une erreur de type I/O est survenue."); 
      SLog.logWriteLine(e.getMessage());
      e.printStackTrace();                                              
    }
      
    SLog.logWriteLine("Message SJFrameRenderer : Fin de l'application."); 
    SLog.logWriteLine("-----------------------------------------------");
    SLog.logWriteLine();
  }
  
  /**
   * Action permettant de dfinir le nom du fichier en lecture.
   */
  private void actionBrowser()
  {
    JFileChooser jfc = new JFileChooser(System.getProperty("user.dir"));

    int returnValue = jfc.showOpenDialog(null);
    
    if (returnValue == JFileChooser.APPROVE_OPTION) 
    {
      File selectedFile = jfc.getSelectedFile();
      this.textFieldFileName.setText(selectedFile.getName());  
    }
  }
  
}
