/**
 * 
 */
package sim.geometry;

import java.io.BufferedWriter;
import java.io.IOException;

import sim.exception.SConstructorException;
import sim.exception.SNoImplementationException;
import sim.graphics.SPrimitive;
import sim.math.SVector3d;
import sim.math.SVectorUV;
import sim.readwrite.SKeyWordDecoder;
import sim.util.SBufferedReader;
import sim.util.SInitializationException;
import sim.util.SStringUtil;

/**
 * La classe </b>SConeGeometry</b> reprsentant la gomtrie d'un cne. 
 * La base du cne est situe en son point P1 et la pointe du cne est situe en son point P2. 
 * 
 * @author Simon Vzina
 * @since 2015-08-11
 * @version 2017-12-20 (version labo  Le ray tracer v2.1)
 */
public class SConeGeometry extends STubeGeometry {

  //-------------
  // VARIABLES //
  //-------------
  
  /**
   * La variable <b>disk</b> correspond au disque situ  la base du cne.
   */
  private SDiskGeometry disk; 
  
  //---------------------------
  // Paramtres de prcalcul //
  //---------------------------
  
  /**
   * La variable <b>H</b> correspond  la hauteur du cne. 
   */
  private double H;           
 
  //-----------------
  // CONSTRUCTEURS //
  //-----------------
  
  /**
   * Construction d'un cne par dfaut. 
   */
  public SConeGeometry()
  {
    this(DEFAULT_P1, DEFAULT_P2, DEFAULT_R);
  }

  /**
   * Constructeur d'un cne.
   * 
   * @param P1 - Le point de la base du cne. 
   * @param P2 - Le point su sommet du cne.
   * @param R - Le rayon de la base du cne.
   */
  public SConeGeometry(SVector3d P1, SVector3d P2, double R)
  {
    this(P1, P2, R, null);
  }

  /**
   * Constructeur d'un cne avec une primitive comme parent.
   * 
   * @param P1 - Le point de la base du cne. 
   * @param P2 - Le point su sommet du cne.
   * @param R - Le rayon de la base du cne.
   * @param parent - La primitive en parent.
   * @throws SConstructorException S'il y a eu une erreur lors de la construction de la gomtrie.
   */
  public SConeGeometry(SVector3d P1, SVector3d P2, double R, SPrimitive parent)throws SConstructorException
  {
    super(P1, P2, R, parent);
    
    try{
      initialize();
    }catch(SInitializationException e){
      throw new SConstructorException("Erreur SConeGeometry 001 : Une erreur lors de l'initialisation est survenue." + SStringUtil.END_LINE_CARACTER + "\t" + e.getMessage());
    }
  }
  
  /**
   * Constructeur du cne  partir d'information lue dans un fichier de format txt.
   * Puisqu'une gomtrie est construite  l'intrieure d'une primitive, une rfrence  celle-ci doit tre intgre au constructeur pour y a voir accs.
   *
   * @param sbr - Le BufferedReader cherchant l'information dans le fichier txt.
   * @param parent - La primitive qui fait la construction de cette gomtrie (qui est le parent).
   * @throws IOException Si une erreur de de type I/O est lance.
   * @throws SConstructorException S'il y a eu une erreur lors de la construction de la gomtrie.
   * @see SBufferedReader
   * @see SPrimitive
   */
  public SConeGeometry(SBufferedReader sbr, SPrimitive parent)throws IOException, SConstructorException
  {
    super(sbr, parent);
    
    try{
      initialize();
    }catch(SInitializationException e){
      throw new SConstructorException("Erreur SConeGeometry 002 : Une erreur lors de l'initialisation est survenue." + SStringUtil.END_LINE_CARACTER + "\t" + e.getMessage());
    }
  }

  //------------
  // MTHODES //
  //------------
  
  @Override
  public int getCodeName()
  {
    return SAbstractGeometry.CONE_CODE;
  }

  @Override
  public boolean isClosedGeometry()
  {
    return true;
  }

  @Override
  public SRay intersection(SRay ray) throws SAlreadyIntersectedRayException
  {
    return ray;
  }
    
  /**
   * Mthode permettant d'valuer l'intersection entre un rayon et la partie cnique du cne.
   * 
   * @param ray - Le rayon  intersecter.
   * @return Le rayon avec les caractristiques de l'intersection entre avec la partie cnique du cne.
   */
  private SRay conicIntersection(SRay ray)
  {
    throw new SNoImplementationException("La mthode n'a pas t implmente.");
  }
  
  @Override
  public boolean isInside(SVector3d v)
  {
    return SGeometricUtil.isOnConeSurface(P1, S12, R, H, v) < 0;      
  }
  
  @Override
  protected SVector3d evaluateIntersectionNormal(SRay ray, double intersection_t)
  {
    return SGeometricUtil.outsideTwoInfinitesConesNormal(P1, S12, R, H, ray.getPosition(intersection_t));
  }

  /**
   * Mthode pour faire l'initialisation de l'objet aprs sa construction.
   * 
   * @throws SInitializationException Si une erreur est survenue lors de l'initialisation.
   */
  private void initialize() throws SInitializationException
  {
    H = P2.substract(P1).modulus();                               // hauteur du cne
    
    // Dfinir le disque avec une normale  la surface extrieur au cne.
    disk = new SDiskGeometry(P1, S12.multiply(-1.0), R, null);    
  }

  @Override
  public void write(BufferedWriter bw) throws IOException
  {
    bw.write(SKeyWordDecoder.KW_CONE);
    bw.write(SStringUtil.END_LINE_CARACTER);
    
    //crire les proprits de la classe SCercleGeometry et ses paramtres hrits
    writeSTubeGeometryParameter(bw);
    
    bw.write(SKeyWordDecoder.KW_END);
    bw.write(SStringUtil.END_LINE_CARACTER);
    bw.write(SStringUtil.END_LINE_CARACTER);
  }

 
  @Override
  protected SVectorUV evaluateIntersectionUV(SRay ray, double intersection_t)
  {
    throw new SNoImplementationException("Erreur SConeGeometry 004 : Cette mthode n'a pas encore t implmente.");
  }
  
  @Override
  protected void readingInitialization() throws SInitializationException
  {
    super.readingInitialization();
    
    initialize();
  }
  
  @Override
  public String getReadableName()
  {
    return SKeyWordDecoder.KW_CONE;
  }
  
}//fin de la classe SConeGeometry
