/**
 * 
 */
package sim.geometry;

import java.lang.Comparable;

import sim.exception.SConstructorException;
import sim.exception.SNoImplementationException;
import sim.exception.SRuntimeException;
import sim.math.SAffineTransformation;
import sim.math.SMath;
import sim.math.SMatrix4x4;
import sim.math.SVector3d;
import sim.math.SVectorUV;

/**
 * <p>
 * La classe <b>SRay</b> reprsente un rayon pouvant raliser une intersection avec une gomtrie.
 * Un rayon est un objet <u>immuable</u> qui peut tre dans un tant <b>intersect</b> ou <b>non intersect</b>.
 * </p>
 * 
 * <p>
 * Lorsqu'un rayon est dans un tant <b>intersect</b>, le rayon donne accs  la gomtrie intersecte,
 * qui elle donne accs  une primitive qui donne finalement accs au matriel appliqu sur la gomtrie.
 * C'est le paramtre <i>t</i> reprsentant le temps pour raliser l'intersection qui permet d'ordonner la visibilit des gomtries de la scne.
 * </p>
 * 
 * <p>
 * Cette classe implmente l'interface <b>Comparable</b> (permettant les usages des mthodes de triage des librairies de base de java). 
 * La comparaison sera effectue sur la <b>valeur du champ local t</b> qui reprsente le <b>temps afin d'intersecter une gomtrie</b>. 
 * S'il n'y a <b>pas d'intersection</b>, la valeur de t sera gale  <b>l'infini</b>.
 * </p> 
 *
 * @author Simon Vzina
 * @since 2014-12-30
 * @version 2017-12-20 (version labo  Le ray tracer v2.1)
 */
public class SRay implements Comparable<SRay> {

  //--------------
  // CONSTANTES //
  //--------------
  
  /**
   * La constante <b>MINIMUM_EPSILON</b> correspond  <b>la plus petite valeur du temps mininal de parcours</b> qu'un rayon doit effectuer avant de pouvoir intersecter une gomtrie. 
   * Cette contrainte est ncessaire pour une raison de stabilit numrique. Elle est prsentement dfinie  {@value} elle et doit <b>toujours tre suprieure  SMath.EPSILON</b>.
   */
  public static final double MINIMUM_EPSILON = 1e-8;	
	
  /**
   * La constante <b>MAXIMAM_T</b> correspond  un temps signifiant que le rayon n'a pas russi  intersecter de gomtrie. 
   * Cette valeur correspond  un <b>temps infini</b>.
   */
	private static final double MAXIMUM_T = SMath.INFINITY;	
	
	/**
	 * La constante <b>DEFAULT_REFRACTIV_INDEX</b> correspond  la valeur de l'indice de rfraction par dfaut o un rayon voyage.
	 * Cet indice de rfraction est gal  celui du vide (n = {@value}).
	 */
	public static final double DEFAULT_REFRACTIVE_INDEX = 1.0;			  
	
  //-----------------------
  // VARIABLES STATIQUES //
  //-----------------------
	
	/**
   * La variable <b>epsilon</b> correspond au temps mininal de parcours qu'un rayon doit effectuer avant de pouvoir intersecter une gomtrie.
   */
  private static double epsilon = MINIMUM_EPSILON;    
  
  //-------------
  // VARIABLES //
  //-------------
  
	/**
	 * La variable <b>origin</b> correspond  l'origine du rayon (d'o est lanc le rayon).
	 */
	private final SVector3d origin;		  
	
	/**
	 * La variable <b>direction</b> correspond  la direction du rayon (dans quelle direction le rayon voyagera).
	 * On peut galement comparer cette variable  la vitesse du rayon.
	 */
	private final SVector3d direction;	
	
	/**
	 * La variable <b>as_intersected</b> dtermine si le rayon a dj ralis une intersection.
	 */
	private final boolean as_intersected;	         
	
	/**
	 * La varibale <b>geometry</b> correspond  la gomtrie intersect par le rayon.
	 */
	private final SGeometry geometry;
	
	/**
	 * La variable <b>normal</b> correspond  la normale  la surface sur la gomtrie intersecte  l'endroit o l'intersection a t ralise.
	 * Cette normale correspond  la <u>normale extrieure</u>  la gomtrie. 
	 */
	private final SVector3d normal;			     
	
	/**
	 * La variable <b>uv</b> correspond  la coordonne <i>uv</i> de texture associe  la gomtrie intersecte par le rayon.
	 */
	private final SVectorUV uv;          
	
	/**
	 * La variable <b>int_t</b> correspond au temps requis au rayon pour intersecter la gometrie.
	 */
	private final double int_t;					         
	
	/**
	 * La variable <b>refractive_index</b> correspond  l'indice de rfraction du milieu o voyage le rayon.
	 */
	private final double refractive_index;
	
	/**
	 * La variable <b>previous_ray</b> correspond au rayon prcdent au rayon prsent dans la <i>hirarchie rcursive</i> du processus de lanc de rayon.
	 */
	private final SRay previous_ray;
	
	//-----------------------
	// CONSTRUCTEUR PUBLIC //
	//-----------------------
	
	/**
	 * Constructeur d'un rayon dont l'objectif sera de tenter d'intersecter une gomtrie.
	 * 
	 * @param origin L'origine du rayon.
	 * @param direction La direction du rayon.
	 * @param refractive_index L'indice de rfraction du milieu o voyage le rayon.
	 * @throws SConstructorException Si la direction du rayon correspond  un vecteur nul (pas d'orientation).
	 */
	public SRay(SVector3d origin, SVector3d direction, double refractive_index) throws SConstructorException
	{
	  this(origin, direction, refractive_index, null);
	}
	
	//----------------------
	// CONSTRUCTEUR PRIV //
	//----------------------
	
	/**
	 * Constructeur d'un rayon dont l'objectif sera de tenter d'intersection une gomtrie.
	 * Ce rayon possde cependant un rayon parent prcdent dfinissant l'ordre rcursif du lanc des rayons.
	 * 
	 * @param origin L'origine du rayon.
   * @param direction La direction du rayon.
   * @param refractive_index L'indice de rfraction du milieu o voyage le rayon.
   * @param previous_ray Le rayon prcdent dans la hirarchie rcursive du lanc des rayons.
   * @throws SConstructorException Si un paramtre est invalide lors de la construction du rayon.
	 */
	private SRay(SVector3d origin, SVector3d direction, double refractive_index, SRay previous_ray) throws SConstructorException
	{
	  // Vrifier que nous n'avons pas de valeur nulle pour l'origine et l'orientation.
	  if(origin == null || direction == null)
	    throw new SConstructorException("Erreur SRay 001 : Le rayon est mal dfini, car origin = " + origin + " et direction = " + direction + " .");
    
	  // Vrification de l'orientation du rayon.
    if(direction.modulus() < SMath.EPSILON)
      throw new SConstructorException("Erreur SRay 001 : La direction du rayon gale  " + direction + " possde un module infrieur  " + SMath.EPSILON + " ce qui ne peut pas lui octroyer d'orientation.");
    
    this.origin = origin;
    this.direction = direction;
    
    this.as_intersected = false;
    
    this.geometry = null;
    this.normal = null;
    this.uv = null;
    this.int_t = MAXIMUM_T;   
    
    this.refractive_index = refractive_index;
    
    this.previous_ray = previous_ray;
	}
	
	/**
	 * Constructeur d'un rayon ayant ralis une intersection avec une gomtrie (sans coordonne de texture uv).
	 * 
	 * @param origin L'origine du rayon.
	 * @param direction La direction du rayon.
	 * @param geometry La gomtrie intersecte.
	 * @param normal L'orientation de la normale  la surface.
	 * @param t Le temps pour le rayon afin d'atteindre la gomtrie.
	 * @param refractive_index L'indice de rfraction du milieu o voyage le rayon.
	 * @param previous_ray Le rayon prcdent dans la hirarchie rcursive du lanc des rayons.
	 * @throws SConstructorException Si le temps du rayon est infrieur  <i>epsilon</i>.
	 * @throws SConstructorException Si un paramtre est invalide lors de la construction du rayon.
	 */
	private SRay(SVector3d origin, SVector3d direction, SGeometry geometry, SVector3d normal, double t, double refractive_index, SRay previous_ray) throws SConstructorException
	{
	  this(origin, direction, geometry, normal, null, t, refractive_index, previous_ray);
	}
	
	/**
   * Constructeur d'un rayon ayant ralis une intersection avec une gomtrie avec coordonne de testure uv.
   * 
   * @param origin L'origine du rayon.
   * @param direction La direction du rayon.
   * @param geometry La gomtrie intersecte.
   * @param normal L'orientation de la normale  la surface.
   * @param uv La coordonne de texture. Une valeur <b>null</b> signifie qu'il n'y aura pas de coordonne d'attribue.
   * @param t Le temps pour le rayon afin d'atteindre la gomtrie.
   * @param refractive_index L'indice de rfraction du milieu o voyage le rayon.
   * @param previous_ray Le rayon prcdent dans la hirarchie rcursive du lanc des rayons.
   * @throws SConstructorException Si le temps du rayon est infrieur  <i>epsilon</i>.
   * @throws SConstructorException Si un paramtre est invalide lors de la construction du rayon.
   */
  private SRay(SVector3d origin, SVector3d direction, SGeometry geometry, SVector3d normal, SVectorUV uv, double t, double refractive_index, SRay previous_ray) throws SConstructorException
  {
    // Vrifier que nous n'avons pas de valeur nulle pour l'origine et l'orientation.
    if(origin == null || direction == null || normal == null)
      throw new SConstructorException("Erreur SRay 001 : Le rayon est mal dfini, car origin = " + origin + ", direction = " + direction + " et normal = " + normal + " .");
  
    // Vrification que la gomtrie n'est pas NULL.
    if(geometry == null)
      throw new SConstructorException("Erreur SRay 003 : La gomtrie intersecte ne peut pas tre 'null'.");
  
    // Vrification pour teste numrique
    if(t < epsilon)
      throw new SConstructorException("Erreur SRay 002 : Pour des raisons numriques, le temps de l'intersection t = " + t + " ne peut pas tre infrieur  epsilon = " + epsilon + " ou ngatif.");
    
    // Vrification de l'orientation du rayon
    if(direction.modulus() < SMath.EPSILON)
      throw new SConstructorException("Erreur SRay 003 : La direction du rayon gale  " + direction + " possde un module infrieur  " + SMath.EPSILON + " ce qui ne peut pas lui octroyer d'orientation.");
    
    this.origin = origin;
    this.direction = direction;
    this.geometry = geometry;
    this.normal = normal;
    this.int_t = t;
    this.as_intersected = true;
    this.refractive_index = refractive_index;
    this.previous_ray = previous_ray;
    
    // Le paramtre uv peut tre 'null'. Sous cette condition, il n'y aura pas de coordonne uv d'affectable  l'intersection.
    this.uv = uv;    
  }
  
  //------------
  // MTHODES //
  //------------
  
	/**
	 * Mthode pour dfinir le temps/distance minimal que doit effectuer un rayon afin de valider une intersection.
	 * Cette valeur est ncessaire pour des raisons numriques. Elle doit tre positive et suprieure  un valeur minimal de 1e-6. 
	 * Il peut tre affect en fonction de la position du front clipping plane d'une pyramide de vue (view Frustum).
	 * @param e - La valeur d'epsilon.
	 * @throws SRuntimeException Si la nouvelle valeur d'epsilon est infrieure  la valeur minimale accepte par la classe.
	 * @see MINIMUM_EPSILON
	 */
	public static void setEpsilon(double e)throws SRuntimeException
	{ 	
		//Mettre la valeur d'epsilon positive
		if(e < MINIMUM_EPSILON)
		  throw new SRuntimeException("Erreur SRay 004 : Pour des raisons numriques, la nouvelle valeur d'epsilon '" + e + "' doit tre suprieure  " + MINIMUM_EPSILON + ".");
			
		epsilon = e;
	}
	
	/**
	 * Mthode pour obtenir la valeur d'epsilon prsentement en vigueur pour les calculs d'intersection entre les rayons et les gomtries.
	 * @return La valeur d'epsilon.
	 */
	public static double getEpsilon()
	{ 
	  return epsilon;
	}
	
	/**
	 * Mthode pour obtenir l'origine du rayon.
	 * @return L'origine du rayon.
	 */
	public SVector3d getOrigin()
	{ 
	  return origin;
	}
	
	/**
	 * Mthode pour obtenir la direction du rayon.
	 * @return La direction du rayon.
	 */
	public SVector3d getDirection()
	{ 
	  return direction; 
	}
		
	/**
	 * Mthode pour obtenir la position d'un rayon aprs un dplacement de celui-ci d'un temps <i>t</i>.
	 * @param t Le temps  couler dans le dplacement du rayon.
	 * @return Le vecteur position du rayon aprs dplacement.
	 */
	public SVector3d getPosition(double t)
	{
		// Remplacez cette exception par votre implmentation.
		throw new SNoImplementationException("La mthode n'a pas t implmente");
	}
	
	/**
	 * Mthode pour dterminer si le rayon a effectu une intersection avec une gomtrie.
	 * @return <b>true</b> s'il y a eu une intersection et <b>false</b> sinon. 
	 */
	public boolean asIntersected()
	{ 
	  return as_intersected;
	}
	
	/**
	 * Mthode pour dterminer si le rayon courant possde hirarchiquement un rayon avant lui.
	 * 
	 * @return <b>true</b> si un rayon a t lanc hirarchiement avant celui-ci et <b>false</b> sinon.
	 */
	public boolean asPreviousRay()
	{
	  return previous_ray != null;
	}
	
	/**
	 * Mthode pour dterminer si le rayon a effectu une intersection de l'intrieur de la gomtrie.
	 * <p> 
	 * Cette mthode ne fait que comparer la direction du rayon avec la normale  la surface extrieur de la gomtrie.
	 * Mme si la gomtrie n'est pas ferme (donc pas d'intrieur), cette mthode peut quand mme retourner <b>true</b>. 
	 * @return <b>true</b> s'il y a eu une intersection  l'intrieur de la gomtrie et <b>false</b> sinon. 
	 * @throws SNotIntersectedRayException Si le rayon n'a pas effectu d'intersection.
	 */
	public boolean isInsideIntersection() throws SNotIntersectedRayException
	{
	  if(!as_intersected)
	    throw new SNotIntersectedRayException("Erreur SRay 006 : Le rayon n'a pas effectu d'intersection avec une gomtrie.");
	  
	  // Afin de dterminer si une intersection est ralise par l'intrieur de la gomtrie,
	  // nous comparons l'orientation du rayon avec la normal (dite extrieure).
	  // Si les orientations sont identique (produit scalaire positif),
	  // alors l'intersection est intrieur.
	  return direction.dot(normal) > 0;
	}
		
	/**
	 * Mthode pour obtenir le temps (ou la distance puisque la direction est unitaire) afin de raliser 
	 * une intersection entre le rayon et une gomtrie. Si le rayon n'a pas t intersect, la valeur sera <b> l'infini </b>.
	 * @return Le temps pour effectuer l'intersection (s'il y a intersection) et <b>l'infini</b> s'il n'y a <b>pas eu d'intersection</b>.
	 */
	public double getT()
	{ 
	  return int_t;
	}
		
	/**
	 * Mthode pour obtenir l'indice de rfraction du milieu dans lequel le rayon voyage.
	 * @return l'indice de rfraction du milieu o voyage le rayon.
	 */
	public double getRefractiveIndex()
	{ 
	  return refractive_index;
	}
	
	/**
	 * Mthode pour obtenir la position de l'intersection entre le rayon et la gomtrie.
	 * @return La position de l'intersection
	 * @throws SNotIntersectedRayException S'il n'y a pas eu d'intersection avec ce rayon, le point d'intersection est indtermin.
	 */
	public SVector3d getIntersectionPosition() throws SNotIntersectedRayException
	{
		if(!as_intersected)
		  throw new SNotIntersectedRayException("Erreur SRay 006 : Le rayon n'a pas effectu d'intersection avec une gomtrie.");

	  // S'il y a intersection, la valeur de "t" correspond au temps de l'intersection 
	  return getPosition(int_t);
	}
	
	/**
	 * Mthode pour obtenir la gomtrie qui est en intersection avec le rayon.
	 * @return La gomtrie en intersection.
	 * @throws SNotIntersectedRayException S'il n'y a pas eu d'intersection avec ce rayon, la gomtrie est indtermine.
	 */
	public SGeometry getGeometry() throws SNotIntersectedRayException
	{
		if(!as_intersected)
		  throw new SNotIntersectedRayException("Erreur SRay 006 : Le rayon n'a pas effectu d'intersection avec une gomtrie.");
	  
		return geometry;
	}
	
	/**
	 * Mthode pour obtenir la normale  la surface de la gomtrie en intersection avec le rayon.
	 * Cette normale sera oriente dans le <u>sens oppos  la direction du rayon</u>.
	 * <p>
	 * Cette dfinition de la normale permet d'valuer le bon ct de la surface de la gomtrie lorsqu'il est temps d'valuer l'illumination (shading).
	 * </p>
	 * @return La normale  la surface  l'endroit de l'intersection.
	 * @throws SNotIntersectedRayException S'il n'y a pas eu d'intersection avec ce rayon, la normale est indtermine.
	 */
	public SVector3d getShadingNormal() throws SNotIntersectedRayException
	{
		if(!as_intersected)
		  throw new SNotIntersectedRayException("Erreur SRay 007 : Le rayon n'a pas effectu d'intersection avec une gomtrie.");
	  		
		// Comparer l'orientation de la normale  la surface avec l'orientation du rayon.
		// Retourner l'orientation tel que la normale est dans le sens oppos  la direction du rayon.
		if(direction.dot(normal) < 0.0)
      return normal;                            // orientation oppose, donc intersection venant de l'extrieur
    else
      return normal.multiply(-1.0);             // orientation dans le mme sens, donc intersection venant de l'intrieur
	}

	/**
	 * Mthode pour obtenir la normale  la surface d'une gomtrie en intersection avec le rayon.
	 * Cette normale sera orientw dans le <u>sens extrieur de la surface de la gomtrie</u>. 
	 * <p>
	 * Cette dfinition est exactement celle fixe lors de l'intersection du rayon avec la gomtrie.
	 * </p>
	 * @return La normale  la surface extrieur de la gomtrie intersecte par le rayon. 
	 * @throws SNotIntersectedRayException S'il n'y a pas eu d'intersection avec ce rayon, la normale est indtermine.
	 */
	public SVector3d getOutsideNormal() throws SNotIntersectedRayException
	{
	  if(!as_intersected)
      throw new SNotIntersectedRayException("Erreur SRay 008 : Le rayon n'a pas effectu d'intersection avec une gomtrie.");
    
	  return normal;
	}
	
	/**
	 * Mthode pour obtenir la coordonne <i>uv</i> associe  l'intersection sur la gomtrie. 
	 * 
	 * @return La coordonne uv associ  l'intersection sur la gomtrie. 
	 * S'il n'y a pas de coordonne uv associe  l'intersection, la valeur <b>null</b> sera retourne.
	 * @throws SNotIntersectedRayException S'il n'y a pas eu d'intersection avec ce rayon, la coordonne <i>uv</i> ne peut pas avoir t dtermine.
	 * @throws SRuntimeException Le rayon n'a pas de coordonne de texture d'attribu.
	 */
	public SVectorUV getUV() throws SNotIntersectedRayException, SRuntimeException
	{
	  // Vrifier que le rayon possde une coordonne UV.
	  if(!asUV())
	    throw new SRuntimeException("Erreur SRay 010 : Il n'y a pas de coordonne de texture associe  ce rayon.");
	  
	  return uv;
	}
	
	/**
	 * Mthode pour dterminer si le rayon intersect possde une coordonne de <i>uv</i> de texture.
	 * 
	 * @return <b>true</b> s'il y a une coordonne <i>uv</i> et <b>false</b> sinon.
	 * @throws SNotIntersectedRayException Si le rayon n'a pas t intersect, il ne peut pas y avoir de coordonne <i>uv</i> d'attribue au rayon.
	 */
	public boolean asUV() throws SNotIntersectedRayException
	{
	  // Vrifier que le rayon a t intersect.
	  if(!as_intersected)
	    throw new SRuntimeException("Erreur SRay 010 : Le rayon n'a pas effectu d'intersection avec une gomtrie.");
  
	  // Retourner "vrai" si uv n'est pas "null".
	  return uv != null;
	}
	
	/**
	 * Mthode pour gnrer un rayon intersect  partir d'un rayon lance et de ses caractristiques dfinissant l'intersection.
	 * 
	 * @param geometry La gomtrie qui est en intersection avec le rayon.
	 * @param normal La normale  la surface de la gomtrie intersecte.
	 * @param t Le temps requis pour se rendre au lieu de l'intersection sur la gomtrie.
	 * @return Le rayon avec les caractristiques de l'intersection.
	 * @throws SAlreadyIntersectedRayException S'il y a dj eu une intersection avec ce rayon.
	 */
	public SRay intersection(SGeometry geometry, SVector3d normal, double t) throws SAlreadyIntersectedRayException
	{
	  if(as_intersected)
	    throw new SAlreadyIntersectedRayException("Erreur SRay 011 : Ce rayon ne peut pas se faire intersecter, car il est prsentement dj intersect.");
	  	  
	  // Construire un nouveau rayon intersect avec les mmes caractristique que le rayon courant (sauf pour l'tat d'intersection).
	  return new SRay(this.origin, this.direction, geometry, normal, t, this.refractive_index, this.previous_ray);
	}
	
	/**
   * Mthode pour gnrer un rayon intersect  partir d'un rayon lance et de ses caractristiques dfinissant l'intersection.
   * 
   * @param geometry La gomtrie qui est en intersection avec le rayon.
   * @param normal La normale  la surface de la gomtrie intersecte.
   * @param uv La coordonne uv associe  l'intersection.
   * @param t Le temps requis pour se rendre au lieu de l'intersection sur la gomtrie.
   * @return Le rayon avec les caractristiques de l'intersection.
   * @throws SAlreadyIntersectedRayException S'il y a dj eu une intersection avec ce rayon.
   */
  public SRay intersection(SGeometry geometry, SVector3d normal, SVectorUV uv, double t) throws SAlreadyIntersectedRayException
  {
    if(as_intersected)
      throw new SAlreadyIntersectedRayException("Erreur SRay 012 : Ce rayon ne peut pas se faire intersecter, car il est prsentement dj intersect.");
    
    // Construire un nouveau rayon intersect avec les mmes caractristique que le rayon courant (sauf pour l'tat d'intersection avec coordonne de texture UV).
    return new SRay(this.origin, this.direction, geometry, normal, uv, t, this.refractive_index, this.previous_ray);
  }
  
  /**
   * Mthode pour construire un rayon rcursif au rayon courant. 
   * L'origine du lanc du nouveau rayon correspondra  la position de l'intersection du rayon courant.
   * 
   * @param direction La nouvelle direction du rayon rcursif.
   * @param refractive_index L'indice de rfraction o voyagera le rayon rcursif.
   * @return Un rayon rcursif au rayon courant.
   * @throws SNotIntersectedRayException Si le rayon n'a pas dj ralis une intersection.
   */
  public SRay castRecursiveRay(SVector3d direction, double refractive_index) throws SNotIntersectedRayException
  {
    // Vrifier que le rayon courant a dj ralis une intersection.
    if(!as_intersected)
      throw new SNotIntersectedRayException("Erreur SRay 013 : Ce rayon n'a pas t intersect, il ne peut pas y avoir de lanc rcursif de ce rayon");
    
    // Construire le nouveau rayon rcursif avec le rayon courant comme tant le parent du rayon relanc.
    return new SRay(getIntersectionPosition(), direction, refractive_index, this);
  }
  
  /**
   * Mhode qui effectue la transformation d'un rayon <b>non intersect</b>  partir d'une matrice de transformation linaire de format 4x4.
   * 
   * @param transformation La matrice de transformation.
   * @return Le rayon transform par la matrice de transformation.
   * @throws SAlreadyIntersectedRayException Si le rayon a t intersect, il ne peut pas tre transform.
   */
  public SRay transformNotIntersectedRay(SMatrix4x4 transformation) throws SAlreadyIntersectedRayException
  {
    if(as_intersected)
      throw new SAlreadyIntersectedRayException("Erreur SRay 014 : Puisque ce rayon a t intersect, il ne peut pas tre transform.");
    
    throw new SNoImplementationException("Erreur SRay : La mthode n'est pas implmente.");
   
  }
  
  /**
   * Mhode qui effectue la transformation d'un rayon <b>non intersect</b>  partir d'une matrice de transformation linaire de format 4x4.
   * Cette version utilile le paramtre <i>transformed_axis_origin</i> qui reprsente l'origine du systme d'axe
   * transform par la matrice de transformation. Cette version permet la transformation du rayon avec 
   * moins d'allocation de mmoire. Cette version est pratique si la transformation de l'origine a t
   * pralablement calcul.
   * 
   * @param transformation La matrice de transformation.
   * @param transformed_origin L'origine du systme d'axe transform par la matrice de transformation.
   * @return Le rayon transform par la matrice de transformation.
   * @throws SAlreadyIntersectedRayException Si le rayon a t intersect, il ne peut pas tre transform.
   */
  public SRay transformNotIntersectedRay(SMatrix4x4 transformation, SVector3d transformed_axis_origin) throws SAlreadyIntersectedRayException
  {
    if(as_intersected)
      throw new SAlreadyIntersectedRayException("Erreur SRay 014 : Puisque ce rayon a t intersect, il ne peut pas tre transform.");
        
    throw new SNoImplementationException("Erreur SRay : La mthode n'est pas implmente.");
  }
      
  @Override
	public int compareTo(SRay ray) 
	{
		return Double.compare(this.int_t, ray.int_t);
	}
	
	@Override
  public String toString()
  {
    return  "SRay [origin=" + origin + ", direction=" + direction + 
        ", as_intersected=" + as_intersected + ", geometry=" + geometry +
        ", int_t=" + int_t + ", normal=" + normal + ", uv=" + uv + ", refractive_index=" + refractive_index + "]";

  }

  @Override
  public boolean equals(Object obj)
  {
    if (this == obj) {
      return true;
    }
    
    if (obj == null) {
      return false;
    }
    
    if (!(obj instanceof SRay)) {
      return false;
    }
    
    SRay other = (SRay) obj;
    if (as_intersected != other.as_intersected) {
      return false;
    }
    
    if (direction == null) {
      if (other.direction != null) {
        return false;
      }
    } else if (!direction.equals(other.direction)) {
      return false;
    }
    
    if (geometry == null) {
      if (other.geometry != null) {
        return false;
      }
    } else if (!geometry.equals(other.geometry)) {
      return false;
    }
    
    if (normal == null) {
      if (other.normal != null) {
        return false;
      }
    } else if (!normal.equals(other.normal)) {
      return false;
    }
    
    if (uv == null) {
        if (other.uv != null) {
          return false;
        }
      } else if (!uv.equals(other.uv)) {
        return false;
      }
    
    if (origin == null) {
      if (other.origin != null) {
        return false;
      }
    } else if (!origin.equals(other.origin)) {
      return false;
    }
    
    if (!SMath.nearlyEquals(refractive_index, other.refractive_index)) {
      return false;
    }
    
    if (!SMath.nearlyEquals(int_t, other.int_t)){ 
      return false;
    }
    
    return true;
  }
	
}//fin classe SRay
