/**
 * 
 */
package sim.geometry.space;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import sim.geometry.SGeometry;
import sim.geometry.SRay;
import sim.math.SVector3d;

/**
 * La classe <b>SAbstractGeometrySpace</b> reprsente un espace  gomtrie abstrait.
 * 
 * @author Simon Vzina
 * @since 2015-01-11
 * @version 2016-01-01
 */
public abstract class SAbstractGeometrySpace implements SGeometrySpace{

  //Mode de slection du type d'espace des gomtries
  public static final String[] TYPE_OF_SPACE = {"linear", "voxel", "multi_voxel"};
  public static final int LINEAR = 0;
  public static final int VOXEL = 1;
  public static final int MULTI_VOXEL = 2;
  
  /**
   * La variable <b>intersection_test_count</b> correspond au nombre de tests d'intersection qui ont t raliss
   * depuis la cration d'un espace des gomtries.
   */
  private static int intersection_test_count = 0;
  
  /**
   * La variable <b>geometry_list</b> correspond  la liste complte des gomtries dans l'espace des gomtries.
   */
  protected final List<SGeometry> geometry_list;
  
  /**
   * La variable <b>space_initialized</b> permet de dfinir si l'espace des voxels a t initialis.
   */
  protected boolean space_initialized;
  
  //----------------
  // CONSTRUCTEUR //
  //----------------
  
  /**
   * Constructeur par dfaut d'un espace de gomtrie abstraite.
   */
  public SAbstractGeometrySpace()
  {
    geometry_list = new ArrayList<SGeometry>();
    
    space_initialized = false;
  }
  
  //------------
  // MTHODES //
  //------------
  
  @Override
  public void addGeometry(SGeometry geometry)
  {
    geometry_list.add(geometry);
    
    space_initialized = false;
  }
  
	@Override
	public void addGeometry(List<SGeometry> list)
	{
		for(SGeometry g : list)
			addGeometry(g);
	}

	/**
   * Mthode qui value toutes les intersections entre un rayon et les diffrentes gomtries d'une liste ne dpassant pas une valeur de temps maximal.
   * Les rayons ayant intersects seront retourns dans une <b>liste trier en ordre croissant de temps</b> (du plus prs au plus loig).
   *
   * @param list - La liste des gomtries.
   * @param ray - Le rayon  intersecter avec les gomtries de l'espace.
   * @param t_max - Le temps maximal.
   * @return Une liste des diffrentes intersections avec les primitives en ordre croissant de temps. Si la <b>liste est vide</b>, il n'y a <b>pas eu d'intersection</b>.
   */
  protected List<SRay> intersections(List<SGeometry> list, SRay ray, double t_max)
  {
    // Compter le nombre de tests d'intersection qui seront effectus lors de l'appel de cette mthode
    increaseIntersectionCount(list.size());
    
    // Liste des rayons intersects  gnrer
    List<SRay> return_list = new ArrayList<SRay>();  
    
    // Itration sur l'ensemble des gomtries de la liste
    for(SGeometry g : list)
    {
      // valuer l'intersection entre le rayon et la gomtrie
      SRay result_ray = g.intersection(ray);  
          
      if(result_ray.asIntersected())    //Vrification de l'intersection
        if(result_ray.getT() < t_max)   //Vrification de ne pas dpasser t_max
          return_list.add(result_ray);
    }
        
    //Trier la list en ordre croissant de valeur t (ce que l'implmentation de Comparable fait pour SRay)
    Collections.sort(return_list); 
        
    return return_list;  
  }
  
  /**
   * <p>
   * Mthode pour fusionner deux listes d'intersections en une nouvelle liste.
   * La nouvelle liste des intersections sera <b>trie en ordre croissant de temps</b>.
   * </p>
   * 
   * <p>
   * Il est important de prciser que cette mthode ne vrifie pas que les rayons  
   * l'intrieur des deux listes sont des rayons ayant raliss des intersection.
   * </p>
   * 
   * @param list1 - La premire liste  fusionner.
   * @param list2 - La deuxime liste  fusionner.
   * @return La fusion des deux listes en <b>tri en ordre croissant de temps</b>.
   */
  protected List<SRay> mergeIntersections(List<SRay> list1, List<SRay> list2)
  {
    List<SRay> return_list = new ArrayList<SRay>();  
    
    return_list.addAll(list1);
    return_list.addAll(list2);
    
    Collections.sort(return_list); 
    
    return return_list;
  }
  
  /**
   * <p>
   * Mthode qui value l'intersection la plus prs entre un rayon et les diffrentes gomtries <b>opaque</b> de l'espace ne dpassant pas un certain temps maximal.
   * Une liste des gomtries transparentes sera ordonne en <b>ordre inverse d'apparition</b> (ordre dcroissant du temps).
   * La <b>liste dbutera par une gomtrie opaque</b> s'il y a eu une intersection avec une gomtrie opaque.
   * S'il n'y a pas eu d'intersection avec une gomtrie opaque, elle contiendra uniquement des gomtries transparentes.
   * </p>
   * 
   * <p>
   * Cette mthode sera utile lors du lanc d'un rayon d'ombre o l'intersection avec une gomtrie transparente
   * ne bloquera pas entirement la lumire (ombre partielle).
   * </p>
   *
   * @param list - La liste des gomtries.
   * @param ray - Le rayon  intersecter avec les gomtries de l'espace.
   * @param t_max - Le temps maximal.
   * @return Une liste des gomtries transparentes intersectes en <b>ordre dcroissant</b> dbutant par une gomtrie opaque s'il y a eu lieu. La liste sera <b>vide</b> s'il n'y a <b>aucune intersection</b> sur le temps maximal.
   */
  protected List<SRay> nearestOpaqueIntersection(List<SGeometry> list, SRay ray, final double t_max)
  {
    // Obtenir la liste complte des intersections tries en ordre croissant
    List<SRay> list_intersection = intersections(list, ray, t_max);
    
    // Iterer sur l'ensemble des intersections et conserver celle 
    // o il y a des gomtries transparentes et une seule opaque s'il y en a une
    List<SRay> return_list = new ArrayList<SRay>();
    
    for(SRay r : list_intersection)
    {
      return_list.add(r);   // ajouter cette intersection
      
      // Si la gomtrie intersect n'est pas transparente,
      // ce sera la dernire intersection ajoute  la liste
      // car elle sera la seule intersection opaque de la liste.
      if(!r.getGeometry().isTransparent())
      {
        Collections.reverse(return_list);       // Inverser l'ordre en ordre dcroissant
        return return_list;                     // retourner une liste avec gomtrie opaque
      }
    }
    
    Collections.reverse(return_list);           // Inverser l'ordre en ordre dcroissant
    return return_list;                         // retourner une liste sans gomtrie opaque
    
    // ANCIENNE VERSION ... BEAUCOUP PLUS COMPLIQUE !!!
    
    /*
    SRay nearest_opaque_intersection = ray;                             //l'intersection opaque la plus prs ... initialement, il n'y en a pas !
    
    List<SRay> list_transparent_intersection = new LinkedList<SRay>();  //liste des intersections avec gomtrie transparente
    
    //- Conserver les intersections transparentes 
    //- Garder en gardant en mmoire l'intersection avec la gomtrie opaque la plus prs
    for(SRay r : list_intersection)
      if(r.getGeometry().isTransparent())                  //Si la gomtrie est transparente
        list_transparent_intersection.add(r);
      else                                                 //Si la gomtrie est opaque
        if(r.getT() < nearest_opaque_intersection.getT())     //Si l'on trouve une gomtrie opaque plus prs que celle trouve avant
          nearest_opaque_intersection = r;
    
    //Construire la liste officielle des intersections transparentes en gardant  
    //uniquement les intersections transparentes avant celle qui est opaque et  
    //mettre l'intersection opaque dans la liste s'il y en a eu une.
    //IMPORTANT : Il faut trier la liste en ordre dcroissant par la suite (la gomtrie opaque sera la premire de la liste, si elle est prsente).
    if(nearest_opaque_intersection.asIntersected())       //S'il y a pas eu d'intersection opaque
    { 
      List<SRay> list_return = new LinkedList<SRay>();    //Nouvelle liste  retourner
      
      for(SRay r : list_transparent_intersection)
        if(r.getT() < nearest_opaque_intersection.getT()) //Ajouter les gomtries transparente plus prs que l'intersection opaque
          list_return.add(r);
      
      list_return.add(nearest_opaque_intersection);     //Ajouter l'intersection opaque
      
      Collections.sort(list_return);              //Trier dans l'ordre croissant
      Collections.reverse(list_return);           //Inverser l'ordre en ordre dcroissant
      return list_return;
    }
    else
    {
      Collections.sort(list_transparent_intersection);      //Trier dans l'ordre croissant
      Collections.reverse(list_transparent_intersection);   //Inverser l'ordre en ordre dcroissant
      return list_transparent_intersection;
    }   
    */
  }
  
  /**
   * Mthode qui effectue la fusion de deux listes de rayon ayant ralis le test de l'intersection
   * avec la gomtrie opaque la plus prs. Ces deux listes seront fusionnes, les premires intersections
   * avec une gomtrie transparente seront conserve jusqu' la 1ire gomtrie opaque et cette squence
   * sera finalement inverse.
   * 
   * @param list1 - La 1ire liste  fusionner.
   * @param list2 - La 2ime liste  fusionner.
   * @return Une nouvelle liste dterminant la 1ire gomtrie opaque intersecte.
   */
  protected List<SRay> mergeNearestOpaqueIntersection(List<SRay> list1, List<SRay> list2)
  {
    // Fusionner les deux listes d'intersection en liste trie
    List<SRay> merge_list = mergeIntersections(list1, list2);
    
    // Iterer sur la liste fusionner et ajouter les intersection transparente
    // et garder seulement une seule intersection avec gomtrie opaque
    List<SRay> return_list = new ArrayList<SRay>();
    
    for(SRay r : merge_list)
    {
      return_list.add(r);
      
      if(!r.getGeometry().isTransparent())
      {
        Collections.reverse(return_list);       // Inverser l'ordre en ordre dcroissant
        return return_list;                     // retourner une liste avec gomtrie opaque
      }
    }
    
    Collections.reverse(return_list);           // Inverser l'ordre en ordre dcroissant
    return return_list;                         // retourner une liste sans gomtrie opaque
  }
  
  /**
   * Mthode qui value l'ensemble des gomtries d'une liste o un vecteur v se retrouve  l'intrieur.
   * 
   * @param list - La liste des gomtries.
   * @param v - La position du vecteur.
   * @return La liste des gomtries o le vecteur v se retrouve  l'intrieur.
   */
  protected List<SGeometry> listInsideGeometry(List<SGeometry> list, SVector3d v)
  {
    // Liste des gomtries o le vecteur v se retrouvera  l'intrieur
    List<SGeometry> return_list = new ArrayList<SGeometry>();  
    
    //Itration sur l'ensemble des gomtries 
    if(list != null)
      for(SGeometry g : list)
        if(g.isClosedGeometry())     // la gomtrie doit tre ferme  
          if(g.isInside(v))          // le vecteur v doit se retrouver  l'intrieur de la gomtrie
            return_list.add(g);
          
    return return_list;
  }
  
  @Override
  public void initialize()
  {
    space_initialized = true;
  }
  
  @Override
  public String toString()
  {
    return new String("Nb geometry : " + geometry_list.size());
  }
  
  /**
   * Mthode pour obtenir le nombre de tests d'intersection qui ont t raliss.
   * 
   * @return Le nombre de tests d'intersection calculs.
   */
  public static int getIntersectionTestCount()
  {
    return intersection_test_count;
  }
  
  /**
   * Mthode pour rinitialiser le comptage des tests d'intersection.
   */
  public static synchronized void resetIntersectinonTestCount()
  {
    intersection_test_count = 0;
  }
  
  /**
   * Mthode pour augmenter le nombre de tests d'intersection raliss.
   * Cette mthode sous cette forme est ncessaire afin d'viter des problmes lors d'un usage de cette classe en mode <i>multiprocessing</i>.
   * 
   * @param value - Le nombre de tests d'intersection  ajouter.
   */
  private static synchronized void increaseIntersectionCount(int value)
  {
    intersection_test_count += value;
  }
  
}//fin classe abstraite SAbstractGeometrySpace
