/**
 * 
 */
package sim.geometry.space;

import java.util.ArrayList;
import java.util.List;

import sim.exception.SNoImplementationException;
import sim.geometry.SAbstractGeometry;
import sim.geometry.SCubeGeometry;
import sim.geometry.SDiskGeometry;
import sim.geometry.SGeometry;
import sim.geometry.SSphereGeometry;
import sim.geometry.STorusGeometry;
import sim.geometry.STransformableGeometry;
import sim.geometry.STriangleGeometry;
import sim.geometry.STubeGeometry;
import sim.math.SAffineTransformation;
import sim.math.SMath;
import sim.math.SMatrix4x4;
import sim.math.SVector3d;

/**
 * La classe <b>SBoundingBoxBuilder</b> reprsente un constructeur de bote englobante autour de diffrentes gomtries.
 * 
 * <p>
 * Les gomtries pouvant tre englobes par une botes sont les suivantes :
 * <ul>- STriangleGeometry</ul>
 * <ul>- SBtriangleGeometry</ul>
 * <ul>- SSphereGeometry</ul>
 * <ul>- STubeGeometry</ul>
 * <ul>- SCylinderGeometry</ul>
 * <ul>- SConeGeometry</ul>
 * <ul>- SCubeGeometry</ul>
 * <ul>- SDiskGeometry</ul>
 * <p>
 * 
 * @author Simon Vzina
 * @since 2015-08-04
 * @since 2017-12-20 (version labo  Le ray tracer v2.1)
 */
public class SBoundingBoxBuilder {

  //--------------
  // CONSTANTES //
  //--------------
  
  /**
   * La constante <b>POSITIVE_MULTIPLICATOR</b> correspond  une constante faiblement suprieure  1.0. 
   */
  private static final double POSITIVE_MULTIPLICATOR = 1.0 + 10.0*SMath.EPSILON;
  
  /**
   * La constante <b>POSITIVE_MULTIPLICATOR</b> correspond  une constante faiblement infrieure  1.0. 
   */
  private static final double NEGATIVE_MULTIPLICATOR = 1.0 - 10.0*SMath.EPSILON;
  
  //----------------
  // CONSTRUCTEUR //
  //----------------
  
  /**
   * Constructeur d'un fabriquant de bote englobante par dfaut. 
   */
  public SBoundingBoxBuilder()
  {
    
  }

  //------------
  // MTHODES //
  //------------
  
  /**
   * Mthode pour obtenir une bote englobante pour une gomtrie. 
   * Si la gomtrie n'est pas admissible  recevoir une bote englobante, la rfrence <b>null</b> sera retourne.
   * 
   * @param geometry La gomtrie  tre embote.
   * @return La bote englobante autour de la gomtrie ou <b>null</b> si aucune bote n'a t construite.
   */
  public SBoundingBox buildBoundingBox(SGeometry geometry)
  {
    switch(geometry.getCodeName())
    {
      case SAbstractGeometry.TRIANGLE_CODE : 
      case SAbstractGeometry.BTRIANGLE_CODE :     return buildBoundingBoxForSTriangleGeometry((STriangleGeometry)geometry);
      
      case SAbstractGeometry.TRANSFORMABLE_CODE : return buildBoundingBoxForSTransformableGeometry((STransformableGeometry)geometry);
          
      case SAbstractGeometry.SPHERE_CODE :        return buildBoundingBoxForSSphereGeometry((SSphereGeometry)geometry);
      
      case SAbstractGeometry.TUBE_CODE : 
      case SAbstractGeometry.CYLINDER_CODE :  
      case SAbstractGeometry.CONE_CODE :          return buildBoundingBoxForSTubeGeometry((STubeGeometry)geometry);
      
      case SAbstractGeometry.CUBE_CODE :          return buildBoundingBoxForSCubeGeometry((SCubeGeometry)geometry);
      
      case SAbstractGeometry.DISK_CODE :          return buildBoundingBoxForSDiskGeometry((SDiskGeometry)geometry);
      
      case SAbstractGeometry.TORUS_CODE :         return buildBoundingBoxForSTorusGeometry((STorusGeometry)geometry);
      
      default : return null;
    }
  }
  
  /**
   * Mthode pour obtenir une bote englobante autour de la gomtrie du STriangleGeometry.
   * 
   * @param triangle La gomtrie du triangle.
   * @return La bote englobante autour du triangle.
   */
  private SBoundingBox buildBoundingBoxForSTriangleGeometry(STriangleGeometry triangle)
  {
    SVector3d[] tab = { triangle.getP0(), triangle.getP1(), triangle.getP2() };
    
    SVector3d min = SVector3d.findMinValue(tab);
    SVector3d max = SVector3d.findMaxValue(tab);
    
    //Si le triangle est parallle  l'un des 6 plans de la bote, la bote sera mal dfinie.
    //Modifions par un "delta" nos coordonnes pour viter cette situation  
    return new SBoundingBox(triangle, buildMinValue(min), buildMaxValue(max));
  }
   
  /**
   * Mthode pour obtenir une bote englobante autour de la gomtrie du STransformableGeometry.
   * 
   * @param geometry La gomtrie transformable.
   * @return La bote englobante autour de la gomtrie transformable. Elle sera <b>null</b> si la gomtrie ne peut pas avoir de bote englobante.
   */
  private SBoundingBox buildBoundingBoxForSTransformableGeometry(STransformableGeometry geometry)
  {
    // Gnrer l'ensemble des huit points dsignant la bote englobante
    List<SVector3d> list = generateBoundingBoxListPoint(geometry);
    
    // Si la gomtrie n'est pas supporte, il faut retourner une bote 'null'.
    if(list.isEmpty())
      return null;
    
    return new SBoundingBox(geometry, list);
  }
  
  /**
   * Mthode pour gnrer une liste de huit points dsignant la bote englobante.
   * Ces points seront transforms selon les matrices de transformation de la gomtrie transformable 
   * afin de s'assurer de la bonne dimension de la bote englobante aprs transformation.
   * 
   * @param geometry La gomtrie transformable.
   * @return Une liste de point correspondant  la bote englobante de la gomtrie transformable.
   */
  private List<SVector3d> generateBoundingBoxListPoint(STransformableGeometry geometry)
  {
    // La liste des points avant la transformation
    List<SVector3d> list;
    
    // La gomtrie interne  la gomtrie transformable
    SGeometry inside_geometry = geometry.getGeometry();
    
    // Vrifier s'il y a une gomtrie interne  la gomtrie transformable
    if(inside_geometry == null)
      return new ArrayList<SVector3d>();
    
    // Obtenir la liste des huit points de la bote englobante
    // Il faudra gnrer de faon rcursive pour le cas d'une gomtrie transformable dans une gomtrie transformable.
    // Il faudra gnrer la bote englobante d'une gomtrie quelconque et obenir ses huit points sinon. 
    if(inside_geometry.getCodeName() == SAbstractGeometry.TRANSFORMABLE_CODE)
      list = generateBoundingBoxListPoint((STransformableGeometry)inside_geometry);
    else
    {
      SBoundingBox box = buildBoundingBox(inside_geometry);
      
      // Vrifier si la gomtrie peut avoir une bote englobante, donc bote non null
      if(box == null)
        return new ArrayList<SVector3d>();
      
      list = box.getListPoint();
    }
    
    // Transformer la liste des points dfinissant la bote englobante.
    return transformBoundingBoxVertex(geometry, list);
  }
  
  /**
   * Mthode pour transformer la liste des huits points dfinissant la bote englobante autour d'une gomtrie transformable
   * afin de les apporter dans l'espace monde de la gomtrie.
   * 
   * @param geometry La gomtrie transformable.
   * @param vertex_list La liste des huits sommets de la bote enblobante.
   * @return La liste des huits sommet de la bote englobante dans l'espace monde de la gomtrie.
   */
  private List<SVector3d> transformBoundingBoxVertex(STransformableGeometry geometry, List<SVector3d> vertex_list)
  {
    throw new SNoImplementationException("C'est mthode n'a pas t implmente.");
  }
    
  /**
   * Mthode pour obtenir une bote englobante autour de la gomtrie du SSphereGeometry.
   * 
   * @param sphere La sphre.
   * @return La bote englobante de la sphre.
   */
  private SBoundingBox buildBoundingBoxForSSphereGeometry(SSphereGeometry sphere)
  {
    SVector3d position = sphere.getPosition();
    double positive_R = sphere.getRay();
    double negative_R = -1*positive_R;
    
    SVector3d min = position.add( new SVector3d(negative_R, negative_R, negative_R));
    SVector3d max = position.add( new SVector3d(positive_R, positive_R, positive_R));
    
    return new SBoundingBox(sphere, buildMinValue(min), buildMaxValue(max));
  }
  
  /**
   * Mthode pour obtenir une bote englobante autour de la gomtrie du SDiskGeometry.
   * 
   * @param disk Le disque.
   * @return La bote englobante du disque.
   */
  private SBoundingBox buildBoundingBoxForSDiskGeometry(SDiskGeometry disk)
  {
    // L'implmentation sera identique  cette de la sphre, 
    // car un disque est entirement localis  l'intrieur d'une sphre peut importe l'orientation de sa normale.
    // Ce choix n'est cependant pas optimal !!!
    // Exploiter l'orientation de la normale  la surface pour limiter la taille de la bote serait une bonne ide.
    
    SVector3d position = disk.getPosition();
    double positive_R = disk.getRay();
    double negative_R = -1*positive_R;
    
    SVector3d min = position.add( new SVector3d(negative_R, negative_R, negative_R));
    SVector3d max = position.add( new SVector3d(positive_R, positive_R, positive_R));
    
    return new SBoundingBox(disk, buildMinValue(min), buildMaxValue(max));
  }
  
  /**
   * Mthode pour obtenir une bote englobante autour de la gomtrie du STubeGeometry.
   * 
   * @param tube Le tube.
   * @return La bote englobante du tube.
   */
  private SBoundingBox buildBoundingBoxForSTubeGeometry(STubeGeometry tube)
  {
    SVector3d[] tab = { tube.getP1(), tube.getP2() };
    
    SVector3d min = SVector3d.findMinValue(tab);
    SVector3d max = SVector3d.findMaxValue(tab);
    
    double positive_R = tube.getRay();
    double negative_R = -1*positive_R;
    
    min = min.add( new SVector3d(negative_R, negative_R, negative_R));
    max = max.add( new SVector3d(positive_R, positive_R, positive_R));
    
    return new SBoundingBox(tube, buildMinValue(min), buildMaxValue(max));
  }
  
  /**
   * Mthode pour obtenir une bote englobante autour de la gomtrie du STubeGeometry.
   * 
   * @param tube Le tube.
   * @return La bote englobante du tube.
   */
  private SBoundingBox buildBoundingBoxForSCubeGeometry(SCubeGeometry cube)
  {
    SVector3d position = cube.getPosition();
    
    double positive_R = cube.getSize() / 2.0;
    double negative_R = -1*positive_R;
    
    SVector3d min = position.add( new SVector3d(negative_R, negative_R, negative_R));
    SVector3d max = position.add( new SVector3d(positive_R, positive_R, positive_R));
    
    return new SBoundingBox(cube, buildMinValue(min), buildMaxValue(max));
  }
  
  /**
   * Mthode pour obtenir une bote englobante autour de la gomtrie du STorusGeometry.
   * 
   * @param torus Le tore.
   * @return La bote englobante du tore.
   */
  private SBoundingBox buildBoundingBoxForSTorusGeometry(STorusGeometry torus)
  {
    SVector3d position = torus.getPosition();
        
    double positive_R = torus.getMajorRadius() + 2.0*torus.getMinorRadius(); // on va pas prendre de chance ...
    double negative_R = -1*positive_R;
    
    SVector3d min = position.add( new SVector3d(negative_R, negative_R, negative_R));
    SVector3d max = position.add( new SVector3d(positive_R, positive_R, positive_R));
    
    return new SBoundingBox(torus, buildMinValue(min), buildMaxValue(max));
  }
  
  /**
   * Mthode pour gnrer un vecteur minimal avec un lger dplacement le rendant encore <b>plus petit</b>.
   * 
   * @param min Le vecteur minimal de rfrence.
   * @return Le vecteur minimal lgrement plus petit.
   */
  private SVector3d buildMinValue(SVector3d min)
  {
    double x,y,z;
    
    if(min.getX() < 0)
      x = min.getX() * POSITIVE_MULTIPLICATOR;
    else
      x = min.getX() * NEGATIVE_MULTIPLICATOR;
    
    if(min.getY() < 0)
      y = min.getY() * POSITIVE_MULTIPLICATOR;
    else
      y = min.getY() * NEGATIVE_MULTIPLICATOR;
    
    if(min.getZ() < 0)
      z = min.getZ() * POSITIVE_MULTIPLICATOR;
    else
      z = min.getZ() * NEGATIVE_MULTIPLICATOR;
    
    return new SVector3d(x, y, z);
  }
  
  /**
   * Mthode pour gnrer un vecteur maximal avec un lger dplacement le rendant encore <b>plus grand</b>.
   * 
   * @param min Le vecteur maximal de rfrence.
   * @return Le vecteur minimal lgrement plus petit.
   */
  private SVector3d buildMaxValue(SVector3d max)
  {
    double x,y,z;
    
    if(max.getX() > 0)
      x = max.getX() * POSITIVE_MULTIPLICATOR;
    else
      x = max.getX() * NEGATIVE_MULTIPLICATOR;
    
    if(max.getY() > 0)
      y = max.getY() * POSITIVE_MULTIPLICATOR;
    else
      y = max.getY() * NEGATIVE_MULTIPLICATOR;
    
    if(max.getZ() > 0)
      z = max.getZ() * POSITIVE_MULTIPLICATOR;
    else
      z = max.getZ() * NEGATIVE_MULTIPLICATOR;
    
    return new SVector3d(x, y, z);
  }
  
}//fin de la classe SBoundingBoxBuilder
