/**
 * 
 */
package sim.geometry.space;

import java.util.ArrayList;
import java.util.List;

import sim.math.SVector3d;

/**
 * <p>Classe reprsentant un constructeur de voxel. L'objectif sera de faire correspondre une coordonne espace
 * monde  une coordonne espace voxel. Pour ce faire, il faudra spcifier <b>1 unit voxel = x unit monde</b>.</p> 
 * <p>Prenant en argument un vecteur ou une bote englobante,
 * le constructeur de voxel va convertir en espace voxel les coordonne (x,y,z) des arguments. Dans le cas d'une bote,
 * il est possible que la bote ncessite plusieurs voxels en raison de la grande taille de la bote
 * ou de la petite taille espace du voxel.</p>
 * 
 * @author Simon Vzina
 * @since 2015-08-04
 * @version 2015-08-04
 */
public class SVoxelBuilder {

  private final double dimension;   //la dimension espace d'un unit voxel ("1" unit voxel = "dimension" unit monde)
 
  /**
   * Constructeur d'un constructeur de voxel.
   * @param dimension - La dimension d'un voxel en unit monde.
   */
  public SVoxelBuilder(double dimension)
  {
    this.dimension = dimension;
  }

  /**
   * Mthode pour obtenir la dimension espace d'un voxel ("1" unit voxel = "dimension" unit monde).
   * @return La dimension d'un voxel en espace monde.
   */
  public double getDimension()
  {
    return dimension;
  }
  
  /**
   * Mthode pour construire un voxel dans lequel un vecteur 3d est situ  l'intrieur.
   * @param v - Le vecteur.
   * @return Le voxel o est situ le vecteur 3d.
   */
  public SVoxel buildVoxel(SVector3d v)
  {
    int x_voxel = (int)Math.floor(v.getX()/dimension);
    int y_voxel = (int)Math.floor(v.getY()/dimension);
    int z_voxel = (int)Math.floor(v.getZ()/dimension);
    
    return new SVoxel(x_voxel, y_voxel, z_voxel);
  }
  
  /**
   * Mthode pour construire une liste de voxel dans lequel une bote englobant est situe  l'intrieur.
   * Si la dimension du voxel est grande et que la bote est petite, la liste peut contenir qu'un seul voxel.
   * @param box - La bote.
   * @return La liste de voxels o est situ la bote.
   */
  public List<SVoxel> buildVoxel(SBoundingBox box)
  {
    List<SVoxel> list = new ArrayList<SVoxel>();
    
    SVoxel min = buildVoxel(box.getMinPoint());
    SVoxel max = buildVoxel(box.getMaxPoint());
    
    if(min.equals(max))   //la bote se retrouve dans un seul voxel
      list.add(min);
    else                  //la bote se retouve dans plusieurs voxels
      for(int i = min.getX(); i <= max.getX(); i++)
        for(int j = min.getY(); j <= max.getY(); j++)
          for(int k = min.getZ(); k <= max.getZ(); k++)
            list.add(new SVoxel(i, j, k));
    
    return list;
  }
  
}//fin de la classe SVoxelBuilder
