/**
 * 
 */
package sim.geometry.space;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import sim.exception.SRuntimeException;
import sim.geometry.SGeometry;
import sim.geometry.SRay;
import sim.math.SVector3d;
import sim.util.SLog;

/**
 * <p>
 * La classe <b>SVoxelSpace</b> reprsentant un espace de gomtries distribues dans un espace de voxels (en trois dimensions).
 * Cette organisation permettra d'effecter l'intersection d'un rayon avec un nombre plus limits de gomtries ce qui va acclrer les calculs.
 * </p> 
 * 
 * <p>
 * Distribue dans une grille rgulire de voxel, chaque gomtrie sera localise dans un ou plusieurs voxels 
 * et le lancer d'un rayon parcourera un nombre limit de voxels ce qui ainsi limitera le nombre de tests d'intersection.
 * </p>
 * 
 * @author Simon Vzina
 * @since 2015-08-04
 * @version 2016-04-04
 */
public class SVoxelSpace extends SAbstractVoxelSpace {

  //-------------
  // VARIABLES //
  //-------------
  
  /**
   * La variable <b>voxel_map</b> correspond  la carte des voxels o sont situes des gomtries admettant une bote englobante.
   */
  private final Map<SVoxel, List<SGeometry>> voxel_map;  
  
  /**
   * La variable <b>voxel_builder</b> correspond au constructeur de voxel. 
   * La taille des voxels sera dtermine par un objet de type SVoxelDimensionEvaluator. 
   */
  private SVoxelBuilder voxel_builder;                   
  
  /**
   * La variable <b>absolute_extremum_voxel</b> correspond  un voxel de coordonne extremums en valeur absolue  celle contenue dans la carte de voxel.
   * Ceci permet de connatre la taille maximal de la carte des voxels.
   */
  private SVoxel absolute_extremum_voxel;               
  
  //----------------
  // CONSTRUCTEUR //
  //----------------
  
  /**
   * Constructeur d'un espace de voxel par dfaut. 
   */
  public SVoxelSpace()
  {
    super();
    
    voxel_map = new HashMap<SVoxel, List<SGeometry>>();
    voxel_builder = null;
    absolute_extremum_voxel = null;
  }

  //------------
  // MTHODES //
  //------------
  
  @Override
  public SRay nearestIntersection(SRay ray, double t_max) throws SRuntimeException
  {
    // Vrifier si le rayon a dj intersect une gomtrie
    if(ray.asIntersected())
      throw new SRuntimeException("Erreur SVoxelSpace 003 : Le rayon en paramtre a dj intersect une gomtrie.");
    
    // Vrifier la valeur de t_max est adquate
    if(t_max < 0.0)
      throw new SRuntimeException("Erreur SVoxelSpace 004 : Le temps maximale ne peut pas tre ngative.");
   
    // Vrifier si l'espace a t initialis
    if(!space_initialized)
      throw new SRuntimeException("Erreur SVoxelSpace 005 : L'espace de gomtries de voxel n'a pas t initialis.");
    
    //Rsultat de l'intersection avec la carte de voxel. Sera gale  "ray" s'il y en a pas eu.
    SRay intersection_in_voxel = nearestIntersectionInVoxelMap(ray, t_max);
    
    //Rsultat de l'intersection avec les gomtries hors voxel
    List<SRay> list_intersection_not_in_voxel = intersections(linear_list, ray, t_max);
    
    // Ajouter l'intersection de la carte de voxel  la liste linaire.
    // Cette liste sera ainsi pas vide.
    list_intersection_not_in_voxel.add(intersection_in_voxel);
    
    // Trier la liste pour le dernier lment ajout.
    Collections.sort(list_intersection_not_in_voxel); 
    
    // Retourner le permier lment de la liste (sera sans intersection s'il n'y en a pas eu).
    return list_intersection_not_in_voxel.get(0);
  }

  /**
   * Mthode pour obtenir l'intersection la plus prs entre un rayon et des gomtries situes dans la carte de voxel.
   * @param ray - Le rayon  intersecter avec les gomtries.
   * @param t_max - Le temps maximal.
   * @return Le rayon avec les caractristiques de l'intersection (s'il y en a eu une).
   */
  private SRay nearestIntersectionInVoxelMap(SRay ray, double t_max)
  {
    //Raliser des calculs d'intersection avec les gomtries de la carte uniquement si elle n'est pas vide
    if(!voxel_map.isEmpty())
    {
      //Crer la ligne de voxel  parcourir un  un
      SFastTraversalVoxelAlgorithm line_of_voxel = new SFastTraversalVoxelAlgorithm(ray, t_max, voxel_builder.getDimension(), absolute_extremum_voxel);
      
      //Faire l'itration sur la ligne de voxel depuis l'origine du rayon
      while(line_of_voxel.asNextVoxel())
      {
        SVoxel voxel = line_of_voxel.nextVoxel();
        
        // Faire le test de l'intersection des gomtries situes dans le voxel en cours
        SRay ray_intersection = nearestIntersectionInVoxelMap(ray, t_max, voxel_map, voxel_builder, voxel);
        
        // Retourner le rsultat de l'intersection s'il y en a une, car elle sera ncessairement de plus court temps
        if(ray_intersection.asIntersected())
          return ray_intersection;
      }
    }
    
    // Aucune intersection valide n'a t trouve.
    return ray;
  }
  
  @Override
  public List<SRay> nearestOpaqueIntersection(SRay ray, double t_max) throws SRuntimeException
  {
    // Vrifier si le rayon a dj intersect une gomtrie
    if(ray.asIntersected())
      throw new SRuntimeException("Erreur SVoxelSpace 006 : Le rayon en paramtre a dj intersect une gomtrie.");
    
    // Vrifier la valeur de t_max est adquate
    if(t_max < 0.0)
      throw new SRuntimeException("Erreur SVoxelSpace 007 : Le temps maximale ne peut pas tre ngative.");
   
    // Vrifier si l'espace a t initialis
    if(!space_initialized)
      throw new SRuntimeException("Erreur SVoxelSpace 008 : L'espace de gomtries de voxel n'a pas t initialis.");
   
    // La liste dtermine dans la carte des gomtries
    List<SRay> list_in_voxel = nearestOpaqueIntersectionInVoxelMap(ray, t_max);
    
    // La liste dtermine dans la liste linaire des gomtries
    List<SRay> list_not_in_voxel = nearestOpaqueIntersection(linear_list, ray, t_max);
    
    // La liste fusionne adquatement
    return mergeNearestOpaqueIntersection(list_in_voxel, list_not_in_voxel);    
  }

  /**
   * Mthode pour obtenir la liste des intersections transparente en ordre dcroissant dont la plus loign (premire de la liste) sera une gomtrie opaque s'il y a eu intersection de cette nature.
   * 
   * @param ray - Le rayon  intersecter.
   * @param t_max - Le temps maximal pouvant tre parcouru par le rayon.
   * @return La liste des intersections transparente en odre dcroissant dont le premier lment sera une gomtrie opaque s'il y a eu intersection de cette nature.
   */
  private List<SRay> nearestOpaqueIntersectionInVoxelMap(SRay ray, double t_max)
  {
    List<SRay> return_list = new ArrayList<SRay>();
    
    if(!voxel_map.isEmpty())
    {
      //Crer la ligne de voxel  parcourir un  un
      SFastTraversalVoxelAlgorithm line_of_voxel = new SFastTraversalVoxelAlgorithm(ray, t_max, voxel_builder.getDimension(), absolute_extremum_voxel);
      
      while(line_of_voxel.asNextVoxel())
      {
        SVoxel voxel = line_of_voxel.nextVoxel();
        
        // Obtenir la liste de l'intersection opaque associ au voxel courant
        List<SRay> list = nearestOpaqueIntersectionInVoxelMap(ray, t_max, voxel_map, voxel_builder, voxel);
        
        // Ajouter cette liste  la liste  retourner
        return_list = mergeNearestOpaqueIntersection(return_list, list);
        
        // Retourner cette liste si l'intersection opaque a dj t trouve.
        if(!return_list.isEmpty())
          if(!return_list.get(0).getGeometry().isTransparent())
            return return_list;
      }
    }
    
    // La liste est vide ou elle contient uniquement des gomtries transparentes
    return return_list;
  }
  
  @Override
  public List<SGeometry> listInsideGeometry(SVector3d v)
  {
    // Vrifier que l'initialisation a t complte
    if(!space_initialized)
      throw new SRuntimeException("Erreur SVoxelSpace 009 : L'espace de voxel n'a pas t initialis.");
    
    // Liste des gomtries o le vecteur v sera situ  l'intrieur.
    // Dbutons avec la liste disponible  partir des informations de la carte des voxels.
    List<SGeometry> inside_list = listInsideGeometryInMap(voxel_map, voxel_builder, v);
    
    // Ajouter les gomtries sans bote o le vecteur v s'y retrouve.
    inside_list.addAll(listInsideGeometry(linear_list, v));
        
    return inside_list;
  }

  @Override
  public void initialize()
  {
    SLog.logWriteLine("Message SVoxelSpace : Construction de l'espace des gomtries avec voxel.");
    
    // Sparateur de la collection de gomtrie
    SGeometryCollectionSplitter splitter = new SGeometryCollectionSplitter(geometry_list, SGeometryCollectionSplitter.SPLIT_BOX_AND_NO_BOX);
    
    // Obtenir la liste des gomtries sans bote englobante et l'affecter  la liste linaire
    linear_list = splitter.getNoBoxList();
    
    // Liste des botes englobantes autour des gomtries de la liste.
    // Nous commencons pas une liste vide afin qu'elle soit remplace par celle donne par le "splitter".
    List<SBoundingBox> bounding_box_list = new ArrayList<SBoundingBox>();   
      
    // Vrifier qu'il y a au moins une liste de liste de botes englobantes dans le "splitter".
    // Puisque le type de sparation gnre uniquement une liste, alors nous pourrons l'utiliser et elle ne sera pas vide.
    if(!splitter.getBoundingBoxSplitList().isEmpty())
      bounding_box_list = splitter.getBoundingBoxSplitList().get(0);
        
    // S'assurer que la liste des botes n'est pas vide, sinon il n'y a pas de carte de voxel  construire
    if(!bounding_box_list.isEmpty())
    {
      // Faire l'valuation de la dimension des voxels et construire le gnrateur de voxel
      //-------------------------------------------------------------------------------------------------------------------------------------------------
      //SVoxelDimensionEvaluator evaluator = new SVoxelDimensionEvaluator(bounding_box_list, SVoxelDimensionEvaluator.ONE_FOR_ONE_ALGORITHM);
      //SVoxelDimensionEvaluator evaluator = new SVoxelDimensionEvaluator(bounding_box_list, SVoxelDimensionEvaluator.BIGGEST_AVERAGE_LENGHT_ALGORITHM);
      SVoxelDimensionEvaluator evaluator = new SVoxelDimensionEvaluator(bounding_box_list, SVoxelDimensionEvaluator.MID_AVERAGE_LENGHT_ALGORITHM);
      //SVoxelDimensionEvaluator evaluator = new SVoxelDimensionEvaluator(bounding_box_list, SVoxelDimensionEvaluator.SMALLEST_AVERAGE_LENGHT_ALGORITHM); 
      
      // Gnrateur de voxel pour la carte de voxel en construction
      voxel_builder = new SVoxelBuilder(evaluator.getDimension()); 
            
      // Construire le voxel d'extrme  la nouvelle carte et le mettre  jour  chaque insertion de gomtrie
      absolute_extremum_voxel = new SVoxel(0, 0, 0);
              
      // Iterer sur l'ensemble botes englobante.
      // Intgrer leur voxels attitrs avec leur gomtrie  la carte des voxels.
      // Mettre  jour le voxel d'extrme
      for(SBoundingBox box : bounding_box_list)
        absolute_extremum_voxel = addGeometryToMap(voxel_map, absolute_extremum_voxel, box.getGeometry(), voxel_builder.buildVoxel(box));     
      
      // Messages multiples  afficher
      SLog.logWriteLine("Message SVoxelSpace : Nombre de gomtries dans la carte de voxels : " + bounding_box_list.size() + " gomtries.");
      SLog.logWriteLine("Message SVoxelSpace : Taille des voxels : " + evaluator.getDimension() + " units.");  
      SLog.logWriteLine("Message SVoxelSpace : Nombre de rfrence  des gomtries : " + evaluateNbGeometryReference(voxel_map) + " rfrences.");
      SLog.logWriteLine("Message SVoxelSpace : Nombre moyen de rfrence  des gomtries par voxel : " + evaluateNbGeometryReferencePerVoxel(voxel_map) + " rfrences/voxel.");
      
      SLog.logWriteLine();
    }//fin if
    else
    {
      // Il n'y a pas de bote englobante de disponible pour l'espace avec voxel
      SLog.logWriteLine("Message SVoxelSpace : Aucune gomtrie ne possde de bote englobante! Le choix d'un espace de gomtries en voxel devient infficace.");
     
      voxel_builder = null;   // Il n'y a pas de constructeur de voxel disponible
    }
    
    SLog.logWriteLine("Message SVoxelSpace : Fin de la construction de l'espace des gomtries avec voxel.");
    SLog.logWriteLine();
    
    // Initialisation de l'espace des voxels complte
    space_initialized = true;
  }
  
}//fin de la classe SVoxelSpace
