/**
 * 
 */
package sim.graphics;

import sim.geometry.SRay;
import sim.graphics.shader.SShader;
import sim.math.SVector3d;
import sim.math.SVectorPixel;

/**
 * Classe abstraite reprsentant un raytracer.
 * @author Simon Vzina
 * @since 2015-04-11
 * @version 2015-07-16
 */
public abstract class SAbstractRaytracer implements SRaytracer {

	 protected final SViewFrustum view_frustum; // pyramide de vue du raytracer
	 protected final SShader shader; // shader du raytracer
	 protected final SViewport viewport; // viewport du raytracer
	
	 /**
	   * Constructeur d'un raytracer.
	   * 
	   * @param view_frustum - La pyramide de vue.
	   * @param shader - Le calculateur d'illumination (<i>shader</i>).
	   * @param viewport - La fentre de rendu (<i>viewport</i>).
	   */
	  public SAbstractRaytracer(SViewFrustum view_frustum, SShader shader, SViewport viewport)
	  {
	    this.view_frustum = view_frustum;
	    this.shader = shader;
	    this.viewport = viewport;
	  }
	  
	/* (non-Javadoc)
	 * @see sim.graphics.SRaytracer#raytrace()
	 */
	@Override
	public void raytrace()
	{
		raytrace(nbPixels());
	}

	/* (non-Javadoc)
	 * @see sim.graphics.SRaytracer#nbPixels()
	 */
	@Override
	public int nbPixels() 
	{
		return viewport.getHeight() * viewport.getHeight();
	}

	/**
	   * Mthode pour valuer la couleur  attribuer  un pixel.
	   * 
	   * @param pixel - La coordonne du pixel.
	   * @return La couleur  attribuer au pixel.
	   */
	abstract protected SColor evaluatePixelColor(SVectorPixel pixel);
	
	/**
	   * Mthode pour calculer la couleur associ  un rayon lanc depuis la position de la camra et traversant un pixel de la pyramide de vue. La coordonne du
	   * pixel sera calcule par la pyramide de vue (view frustum) et la couleur sera dtermine par le shader ayant accs  la gomtrie de la scne et des
	   * matriaux qui s'y trouvent.
	   * 
	   * @param pixel - La coordonne du pixel.
	   * @return La couleur calcule par <b>le lanc d'un rayon</b> dans le pixel.
	   */
	  protected SColor raytracePixel(SVectorPixel pixel)
	  {
	    // La coordonne 3d du pixel dans les coordonnes de la scne
	    SVector3d position_pixel = view_frustum.viewportToViewFrustum(pixel);

	    // Direction du rayon normalis
	    SVector3d direction = position_pixel.substract(view_frustum.getCameraPosition()).normalize();
	   
	    // valuer adquatement l'indice de rfraction et reconstruction du rayon adquatement
	    double refractive_index = shader.evaluateRefractiveIndex(position_pixel);
	    
	    // Construire le rayon avec le bon indice de rfraction
	    SRay ray = new SRay(position_pixel, direction, refractive_index); 

	    // valuer la couleur attribue  ce rayon partant de la camra et traversant un pixel du viewport
	    return shader.shade(ray);
	  }
	  
}//fin de la classe SAbstractRaytracer
