/**
 * 
 */
package sim.graphics;

import java.util.ArrayList;
import java.util.List;
import java.io.FileNotFoundException;
import java.io.BufferedWriter;
import java.io.IOException;

import sim.exception.SConstructorException;
import sim.exception.SRuntimeException;
import sim.geometry.SGeometry;
import sim.geometry.SPlaneGeometry;
import sim.geometry.SSphereGeometry;
import sim.graphics.light.SAmbientLight;
import sim.graphics.light.SApertureMaskLight;
import sim.graphics.light.SDirectionalLight;
import sim.graphics.light.SEllipticalApertureLight;
import sim.graphics.light.SLight;
import sim.graphics.light.SLinearApertureLight;
import sim.graphics.light.SPointLight;
import sim.graphics.light.SRectangularApertureLight;
import sim.graphics.material.SBlinnMaterial;
import sim.graphics.material.SBlinnTextureMaterial;
import sim.graphics.material.SMaterial;
import sim.graphics.SModel;
import sim.math.SVector3d;
import sim.readwrite.SAbstractReadableWriteable;
import sim.readwrite.SKeyWordDecoder;
import sim.util.SBufferedReader;
import sim.util.SInitializationException;
import sim.util.SLog;
import sim.util.SReadingException;
import sim.util.SStringUtil;

/**
 * La classe <b>SScene</b> reprsente une scne en trois dimensions. 
 * Une scne comprend des primitives, des matriaux, des lumires, ainsi que des paramtres de visionnement comme une camra, un viewport et le ray tracer.
 *  
 * @author Simon Vzina
 * @since 2014-12-26
 * @version 2016-05-20
 */
public class SScene extends SAbstractReadableWriteable {

  //--------------
  // CONSTANTES //
  //--------------
  
  /**
   * La constante <b>KEYWORD_PARAMETER</b> correspond  un tableau contenant l'ensemble des mots cls 
   *  utiliser reconnus lors de la dfinition de l'objet par une lecture en fichier.
   */
  private static final String[] KEYWORD_PARAMETER = {
    SKeyWordDecoder.KW_CAMERA, SKeyWordDecoder.KW_VIEWPORT, SKeyWordDecoder.KW_RAYTRACER,
    SKeyWordDecoder.KW_PRIMITIVE, SKeyWordDecoder.KW_BLINN_MATERIAL, SKeyWordDecoder.KW_TEXTURE_MATERIAL, 
    SKeyWordDecoder.KW_AMBIENT_LIGHT, SKeyWordDecoder.KW_DIRECTIONAL_LIGHT, SKeyWordDecoder.KW_POINT_LIGHT,
    SKeyWordDecoder.KW_LINEAR_APERTURE_LIGHT, SKeyWordDecoder.KW_RECTANGULAR_APERTURE_LIGHT,
    SKeyWordDecoder.KW_ELLIPTICAL_APERTURE_LIGHT,SKeyWordDecoder.KW_APERTURE_MASK_LIGHT,
    SKeyWordDecoder.KW_MODEL     
  };
  
  /**
   * La constante <b>DEFAULT_SCENE_FILE_NAME</b> correspond au nom du fichier de scne par dfaut tant gal  {@value}. 
   */
  public static final String DEFAULT_SCENE_FILE_NAME = "default_scene.txt";
  
  //-------------
  // VARIABLES //
  //-------------
  
  /**
   * La variable 'camera' correspond  la camra qui sera utilise pour visualiser la scne.
   */
	private SCamera camera;					                 
	
	/**
	 * La variable 'viewport' correspond  l'cran de vue o seront crit les diffrents pixels de couleur calculs par le ray tracer.
	 */
	private SViewport viewport;		
	
	/**
	 * La variable 'raytracer_builder' correspond  un constructeur du raytracer. Selon ses paramtres, un ray tracer particulier sera ventuellement construit.
	 */
	private SRaytracerBuilder raytracer_builder;     
			
	/**
	 * La variable 'primitive_list' correspond  la liste des primitives situes dans la scne. Une primitive est une gomtrie affecte  un matriel.
	 */
	private final List<SPrimitive> primitive_list;   
	
	/**
	 * La variable 'material_list' correspond  la liste des diffrents matriaux disponibles lors du calcul du rendu.
	 */
	private final List<SMaterial> material_list;		 
	
	/**
	 * La variable 'light_list' correspond  la liste des sources de lumires disponibles pour faire les calculs d'illumination de la scne lors du ray tracing.
	 */
	private final List<SLight> light_list;				   
	
	/**
	 * La variable 'model_list' correspond  la liste des diffrents modles situs dans la scne. Un modle comprend plusieurs primitives (gomtrie + matriel).
	 */
	private final List<SModel> model_list;           
		
	/**
	 * La variable 'geometry_list' correspond  la liste des diffrentes gomtries situs dans la scne.
	 */
	private final List<SGeometry> geometry_list;     
	
	//----------------
	// CONSTRUCTEUR //
	//----------------
	
	/**
	 * Constructeur avec lecture d'un fichier txt.
	 * 
	 * @param file_name - Le nom du fichier en lecture.
	 * @throws FileNotFoundException Si le fichier de lecture n'a pas t trouv.
	 * @throws IOException Si une erreur de type I/O est survenue.
	 * @throws SConstructorException Si une erreur est survenue lors de la construction.
	 */
	public SScene(String file_name) throws FileNotFoundException, IOException, SConstructorException
	{
	  camera = new SCamera();                             
    viewport = new SViewport();                         
    raytracer_builder = new SRaytracerBuilder();        
        
    primitive_list = new ArrayList<SPrimitive>();  
    material_list = new ArrayList<SMaterial>();     
    light_list = new ArrayList<SLight>();              
    model_list = new ArrayList<SModel>();              
      
    geometry_list = new ArrayList<SGeometry>();   
		
    // Choix de la cration de la scne
    if(file_name.equals(DEFAULT_SCENE_FILE_NAME))
      buildDefaultScene();                            // scne par dfaut
    else
      read(file_name);                                // scne en lecture  
	}
		
	//------------
	// MTHODES //
	//------------
	
	/**
	 * Mthode pour obtenir la camra de la scene.
	 * 
	 * @return La camra de la scene.
	 */
	public SCamera getCamera()
	{ 
	  return camera;
	}
	
	/**
	 * Mthode pour obtenir le viewport de la scene.
	 * 
	 * @return Le viewport de la scene.
	 */
	public SViewport getViewport()
	{ 
	  return viewport;
	}
	
	/**
	 * Mthode pour obtenir la liste des primitives de la scne.
	 * 
	 * @return La liste des primitives.
	 */
	public List<SPrimitive> getPrimitiveList(){ return primitive_list; }
	
	/**
	 * Mthode pour obtenir la liste des sources de lumire de la scne.
	 * 
	 * @return La liste des sources de lumire.
	 */
	public List<SLight> getLightList(){ return light_list; }
	
	/**
	 * Mthode pour faire la construction du raytracer associ  la scne en fonction de ses paramtres comme le viewport, la camra, les gomtries et les lumires.
	 * 
	 * @return Le ray tracer de la scne.
	 */
	public SRaytracer buildRaytracer()
	{
	  return raytracer_builder.buildRaytracer(viewport, camera, geometry_list, light_list);
	}

	@Override
	public void write(BufferedWriter bw) throws IOException
	{
		writeComment(bw, "Paramtres du rendu");
		
	  camera.write(bw);
		viewport.write(bw);
		raytracer_builder.write(bw);
			
		writeComment(bw, "Modles de la scne");
		    
		//crire les modle
    for(SModel m : model_list)
      m.write(bw);
    
    writeComment(bw, "Primitives de la scne");
       
    //crire les primitives
    for(SPrimitive p : primitive_list)
      p.write(bw);
    
    writeComment(bw, "Matriaux de la scne");
    
   	//crire les matriaux
		for(SMaterial m : material_list)
			m.write(bw);
		
		writeComment(bw, "Sources de lumire de la scne");
				
		//crire les lumires
		for(SLight l : light_list)
			l.write(bw);		
	}
	
	/**
   * Mthode pour faire l'initialisation de l'objet aprs sa construction.
   * 
   * @throws SInitializationException Si une erreur est survenue lors de l'initialisation.
   */
  private void initialize() throws SInitializationException
	{
		int primitive_count = 0;
		
		//Remplir l'espace de gomtrie avec les gomtries des primitives disponibles.
		//Les primitives sans gomtrie (qui sont  null) ne seront pas ajoutes  l'espace des gomtries.
		for(SPrimitive p : primitive_list)
			if(p.getGeometry() != null)							         //Vrification de la prsence d'une gomtrie
			{
			  geometry_list.add(p.getGeometry());	//Ajout de la gomtrie
			  primitive_count++;
			}
	
		//Mettre l'ensemble des gomtries des modle dans l'espace des gomtries
		for(SModel m : model_list)
		  for(SPrimitive p : m.getPrimitiveList())
		    if(p.getGeometry() != null)                     //Vrification de la prsence d'une gomtrie
		    {
		      geometry_list.add(p.getGeometry());  //Ajout de la gomtrie
		      primitive_count++;
		    }
		
		//Faire l'assignation des matriaux aux primitives
		SMaterialAffectation m_affectation = new SMaterialAffectation(primitive_list, material_list);
		m_affectation.affectation();
		
		//Message du comptage des primitives
    SLog.logWriteLine("Message SScene : La scne comprend " + primitive_count + " primitives.");
	}
	
	@Override
	protected boolean read(SBufferedReader sbr, int code, String remaining_line)throws SReadingException, IOException
	{
		try{
  	  switch(code)
  		{
  			case SKeyWordDecoder.CODE_CAMERA :				        camera = new SCamera(sbr); return true;
  															
  			case SKeyWordDecoder.CODE_VIEWPORT :			        viewport = new SViewport(sbr); return true;
  											
  			case SKeyWordDecoder.CODE_RAYTRACER :             raytracer_builder = new SRaytracerBuilder(sbr); return true;
  			
  			case SKeyWordDecoder.CODE_PRIMITIVE :			        primitive_list.add(new SPrimitive(sbr)); return true;
  															
  			case SKeyWordDecoder.CODE_BLINN_MATERIAL:		      material_list.add(new SBlinnMaterial(sbr)); return true;
  			
  			case SKeyWordDecoder.CODE_TEXTURE_MATERIAL :      material_list.add(new SBlinnTextureMaterial(sbr)); return true;
  			
  			case SKeyWordDecoder.CODE_AMBIENT_LIGHT :		      light_list.add(new SAmbientLight(sbr));	return true;
  															
  			case SKeyWordDecoder.CODE_DIRECTIONAL_LIGHT :	    light_list.add(new SDirectionalLight(sbr));	return true;
  															
  			case SKeyWordDecoder.CODE_POINT_LIGHT :			      light_list.add(new SPointLight(sbr));	return true;
  			
  			case SKeyWordDecoder.CODE_LINEAR_APERTURE_LIGHT : light_list.add(new SLinearApertureLight(sbr)); return true;
  			
  			case SKeyWordDecoder.CODE_RECTANGULAR_APERTURE_LIGHT : light_list.add(new SRectangularApertureLight(sbr)); return true;
        
  			case SKeyWordDecoder.CODE_ELLIPTICAL_APERTURE_LIGHT : light_list.add(new SEllipticalApertureLight(sbr)); return true;
  			
  			case SKeyWordDecoder.CODE_APERTURE_MASK_LIGHT : light_list.add(new SApertureMaskLight(sbr)); return true;
        
  			
  			case SKeyWordDecoder.CODE_MODEL :   SModelReader m_reader = new SModelReader(sbr);
  			
  			                                    if(m_reader.asRead())
  			                                      model_list.add((SModel)m_reader.getValue());
  			                                    
  			                                    return true;
  			
  			default : return false;
  		}
		}catch(SConstructorException e){
		  throw new SReadingException("Erreur SScene 004 : La construction d'un lment de la scne est impossible." + SStringUtil.END_LINE_CARACTER + "\n" + e.getMessage());
		}
	}
			
	@Override
  protected void readingInitialization() throws SInitializationException
  {
    initialize();
  }
	
	@Override
  public String getReadableName()
  {
	  return SKeyWordDecoder.KW_SCENE;
  }
  
  @Override
  public String[] getReadableParameterName()
  {
    String[] other_parameters = super.getReadableParameterName();
    
    return SStringUtil.merge(other_parameters, KEYWORD_PARAMETER);
  }
  
  /**
   * Mthode pour faire le chargement de la scne par la lecture d'un fichier txt.
   * 
   * @param file_name Le nom du fichier txt.
   * @throws FileNotFoundException Si le ficher de lecture n'est pas trouv.
   * @throws SConstructorException S'il y a eu une erreur lors de la construction de l'objet.
   * @throws IOException Si une erreur de type I/O est survenue.
   */
  /*
  private void readScene(String file_name) throws FileNotFoundException, SConstructorException, IOException
  {
    // Trouver le fichier  partir du rpertoire o l'excution de l'application est ralise
    SFileSearch search = new SFileSearch("", file_name);
    
    if(!search.isFileFound())
      throw new FileNotFoundException("Erreur SScene 001 : Le fichier '" + file_name + "' n'est pas trouv.");
    
    if(search.isManyFileFound())
      throw new SConstructorException("Erreur SScene 002 : Le fichier '" + file_name + "' a t trouv plus d'une fois dans les diffrents sous-rpertoires. Veuillez en garder qu'une seule version.");
    
    // Lecture de la scne  partir d'un fichier
    FileReader fr = new FileReader(search.getFileFoundList().get(0));
    SBufferedReader sbr = new SBufferedReader(fr);
    
    try{
      read(sbr);
    }catch(SInitializationException e){
      throw new SConstructorException("Erreur SScene 003 : Une erreur  l'initialisation est survenue." + SStringUtil.END_LINE_CARACTER + "\t" + e.getMessage(), e);
    }
      
    sbr.close();
    fr.close();
  }
  */
  
  /**
   * Mthode pour faire la construction d'une scne par dfaut.
   */
  private void buildDefaultScene()
  {
    SPlaneGeometry plan = new SPlaneGeometry();
    SSphereGeometry sphere = new SSphereGeometry(new SVector3d(0.0, 0.0, 1.0), 0.5);
    
    SPrimitive p1 = new SPrimitive(plan);
    SPrimitive p2 = new SPrimitive(sphere);
    
    SDirectionalLight l1 = new SDirectionalLight();
    
    primitive_list.add(p1);
    primitive_list.add(p2);
    light_list.add(l1);
    
    camera = new SCamera(new SVector3d(-2.0, 0.0, 2.0), new SVector3d(0.0, 0.0, 0.0), new SVector3d(0.0, 0.0, 1.0));
    
    try{
      initialize();
    }catch(SInitializationException e){
      throw new SRuntimeException("Erreur SScene 005 : Une erreur d'initialisation est survenue durant la construction de la scne par dfaut." + SStringUtil.END_LINE_CARACTER + "\t" + e.getMessage(), e);
    }
  }
  
}//fin de la classe SScene
