/**
 * 
 */
package sim.graphics.light;


import java.util.List;
import java.util.LinkedList;

import sim.exception.SNoImplementationException;
import sim.exception.SRuntimeException;
import sim.geometry.SRay;
import sim.geometry.space.SGeometrySpace;
import sim.graphics.SColor;
import sim.graphics.material.SMaterial;
import sim.graphics.shader.SIllumination;
import sim.math.SVector3d;

/**
 * La classe <b>SShadewRay</b> reprsente un rayon d'ombrage associ  une source de lumire clairant un point d'intersection.
 * 
 * @author Simon Vzina
 * @since 2015-01-21
 * @version 2017-08-22
 */
public class SShadowRay {

	final private boolean is_in_shadow;			//dtermine si la surface claire est dans l'ombre de la source de lumire.
	final private SColor filtred_light_color;	//couleur de la source de lumire  l'endroit de l'intersection (ombrage, lumire filtre ou source de lumire intacte).
	
	/**
	 * Constructeur d'un rayon d'ombre associ  une source de lumire.
	 * @param intersection_ray - Le rayon donnant accs  la surface intersecte qui est claire.
	 * @param light - La source de lumire clairant la surface intersecte.
	 * @param geometry_space - L'espace des gomtries de la scne.
	 * @throws SRuntimeException - Si le rayon n'a pas intersect de gomtrie dans l'espace des gomtries.
	 */
	public SShadowRay(SRay intersection_ray, SLight light, SGeometrySpace geometry_space)throws SRuntimeException
	{
		if(!intersection_ray.asIntersected())
			throw new SRuntimeException("Erreur SShadowRay 001 : Le rayon n'a pas effectu d'intersection avec une gomtrie.");
		
		filtred_light_color = evaluateFilteredLightColor(intersection_ray, light, geometry_space);
		is_in_shadow = evaluateIsInShadow(filtred_light_color);
	}
	
	/**
	 * Mthode pour dterminer si la surface d'intersection est situe dans l'ombre
	 * de la source de lumire associe  ce rayon d'ombre.
	 * @return <b>true</b> si la surface intersecte est dans l'ombre de la source de lumire et <b>false</b> sinon.
	 */
	public boolean isInShadow()
	{ 
	  return is_in_shadow;
	}
	
	/**
	 * Mthode pour obtenir la couleur de la source de lumire aprs filtrage caus par son passage au travers des matriaux transparents. 
	 * @return la <b>couleur d'origine</b> de la lumire si aucune gomtrie tait entre la surface intersecte et la source de lumire,
	 * la <b>couleur noire</b> s'il y a eu une gomtrie opaque entre la surface intersecte et la source de lumire (mthode isInShadow() == true) et
	 * une <b>couleur filtre</b> s'il y a eu des gomtries transparentes entre la surface intersecte et la source de lumire. 
	 */
	public SColor filtredLight()
	{ 
	  return filtred_light_color;
	}
	
	/**
	 * Mthode qui value si la surface intersecte est dans l'ombre de la source de lumire ou pas et dtermine
	 * la couleur filtre de la source de lumire si elle se rend  la surface intersecte. 
	 * @param intersection_ray -  Le rayon donnant accs  la surface intersecte  clairer.
	 * @param light - La source de lumire clairant la surface intersecte.
	 * @param orientation_light - L'orientation de la source de lumire.
	 * @param distance - La distance entre la source de lumire et la surface intersecte.
	 * @param geometry_space - L'espace des gomtries de la scne.
	 */
	private SColor evaluateFilteredLightColor(SRay intersection_ray, SLight light, SGeometrySpace geometry_space)
	{
		//Cas de la source de lumire ambiante
		if(light instanceof SAmbientLight)
			return light.getColor();
		
		SVector3d d;		  //orientation de la source de source de lumire
		double distance;	//distance entre la source et la source de lumire
		
		//Cas de la source de lumire directionnelle
		if(light instanceof SDirectionalLight)
		{
			d = ((SDirectionalLight)light).getOrientation();
			distance = Double.POSITIVE_INFINITY;
		}
		else
			//Cas de la source de lumire ponctuelle
		  /*
		  // LIGNE CHANGE POUR PERMETTRE AU SInterferenceLight d'tre inclus ici !
		  if(light instanceof SPointLight)         
			{
				SPointLight point_light = (SPointLight)light;
				
				d = point_light.getOrientation(intersection_ray.getIntersectionPosition());
				distance = point_light.getPosition().substract(intersection_ray.getIntersectionPosition()).modulus();
			}
			*/
		  if(light instanceof SAttenuatedLight)         
      {
		    SAttenuatedLight point_light = (SAttenuatedLight)light;
        
        d = point_light.getOrientation(intersection_ray.getIntersectionPosition());
        distance = point_light.getPosition().substract(intersection_ray.getIntersectionPosition()).modulus();
      }
			else
				throw new SNoImplementationException("Erreur SShadowRay 002 : La source de lumire n'a pas t implmente.");
		  
		SVector3d L = d.multiply(-1.0);		//orientation inverse de la lumire
			
		// Tester si la normale est oriente dans le mme sens que la source de lumire 
		if(intersection_ray.getShadingNormal().dot(L) <= 0.0)
			return SIllumination.NO_ILLUMINATION;
		
		
			//Rayon de l'intersection  la source de lumire (avec indice de rfraction de 1.0, car il n'y aura pas de calcul de rfraction pour ce rayon)
			SRay intersection_to_light_ray = new SRay(intersection_ray.getIntersectionPosition(), L, SRay.DEFAULT_REFRACTIVE_INDEX);
			
			//Liste des gomtries rencontres avant d'atteindre la source de lumire 
			//en ordre inverse et dbutant par une gomtrie opaque s'il y a eu intersection de ce type
			List<SRay> list_transparent = geometry_space.nearestOpaqueIntersection(intersection_to_light_ray, distance);
			
			//Test de la liste vide, il n'y a pas d'intersection, donc pas d'ombre ni lumire altre par des effets de transparence
			if(list_transparent.isEmpty())
				return light.getColor();
			else
			{
				//Si la liste possde une gomtrie opaque (non transparente) au dbut, la lumiere sera bloque
				if(!list_transparent.get(0).getGeometry().isTransparent())
					return SIllumination.NO_ILLUMINATION;
				else
				{
					List<SMaterial> material_list = new LinkedList<SMaterial>();
					
					for(SRay r : list_transparent)
						material_list.add(r.getGeometry().getPrimitiveParent().getMaterial());
						
					return SIllumination.filteredTransparencyLight(material_list, light.getColor());
				}
			}
			
	}
	
	/**
	 * Mthode qui dtermine si une couleur passe en paramtre est considre commme tant dans l'ombre.
	 * @param c - La couleur  dterminer si elle est dans l'ombrage.
	 * @return <b>true</b> si la couleur est 'noir' (0.0, 0.0, 0.0) et <b>false</b> sinon.
	 */
	private boolean evaluateIsInShadow(final SColor c)
	{
		if(c.equals(SIllumination.NO_ILLUMINATION))
			return true;
		else
			return false;
	}
	
}//fin classe SShadowRay
