/**
 * 
 */
package sim.graphics.material;

import sim.graphics.SColor;
import sim.math.SVectorUV;
import sim.readwrite.SWriteable;

/**
 * L'interface SMaterial reprsente un matriel avec diffrentes proprits. 
 * 
 * @author Simon Vzina
 * @since 2015-01-13
 * @version 2015-10-31
 */
public interface SMaterial extends SWriteable {

	/**
	 * Mthode pour obtenir le nom du matriel.
	 * 
	 * @return Le nom du matriel.
	 */
	public String getName();
	
	/**
	 * Mthode pour obtenir la <b>couleur rflchie</b> de faon <u>ambiante</u> par la surface du matriel.
	 * Cette couleur peut tre calcule  partir du produit de la couleur de la surface <b><i>S</i></b> avec la constant
	 * de rflexion ambiant <b><i>k</i>a</b> tel que <ul><b><i>S</i>a = <i>k</i>a*<i>S</i></b>.</ul>
	 * 
	 * @return La couleur de la lumire <b><i>S</i>a</b> rflchie de faon ambiante par la surface.
	 */
	public SColor ambientColor();
	
	/**
   * Mthode pour obtenir la <b>couleur rflchie</b> de faon <u>ambiante</u> par la surface du matriel
   *  partir d'une <b>coordonne <i>uv</i></b> d'une <b>texture</b>. S'il n'y a <b>pas de texture en mmoire</b> ou 
   * que cette <b>mthode est non dfinie</b> par le matriel, la couleur retourne sera celle de la mthode <b>ambiantColor()</b>.
   * 
   * @param uv - La coordonne <i>uv</i> de la texture. 
   * @return La couleur de rflexion ambiante de la texture (si elle est disponible).
   */
  public SColor ambientColor(SVectorUV uv);
    
	/**
	 * Mthode pour obtenir la <b>couleur rflchie</b> de faon <u>diffuse</u> par la surface du matriel.
	 * Cette couleur peut tre calcule  partir du produit de la couleur de la surface <b><i>S</i></b> avec la constant
	 * de rflexion diffuse <b><i>k</i>b</b> tel que <ul><b><i>S</i>d = <i>k</i>d*<i>S</i></b>.</ul>
	 * 
	 * @return La couleur de la lumire <b><i>S</i>d</b> rflchie de faon diffuse par la surface.
	 */
	public SColor diffuseColor();
	
	/**
   * Mthode pour obtenir la <b>couleur rflchie</b> de faon <u>diffuse</u> par la surface du matriel
   *  partir d'une <b>coordonne <i>uv</i></b> d'une <b>texture</b>. S'il n'y a <b>pas de texture en mmoire</b> ou 
   * que cette <b>mthode est non dfinie</b> par le matriel, la couleur retourne sera celle de la mthode <b>diffuseColor()</b>.
   * 
   * @param uv - La coordonne <i>uv</i> de la texture. 
   * @return La couleur de rflexion diffuse de la texture (si elle est disponible).
   */
  public SColor diffuseColor(SVectorUV uv);
  
	/**
	 * Mthode pour obtenir la <b>couleur rflchie</b> de faon <u>spculaire</u> par la surface du matriel.
	 * Cette couleur peut tre calcule  partir du produit de la couleur de la surface <b><i>S</i></b> avec la constant
	 * de rflexion spculaire <b><i>k</i>s</b> tel que <ul><b><i>S</i>s = <i>k</i>s*<i>S</i></b>.</ul>
	 * 
	 * @return La couleur de la lumire <b><i>S</i>s</b> rflchie de faon spculaire par la surface.
	 */
	public SColor specularColor();
	
	/**
   * Mthode pour obtenir la <b>couleur rflchie</b> de faon <u>spculaire</u> par la surface du matriel
   *  partir d'une <b>coordonne <i>uv</i></b> d'une <b>texture</b>. S'il n'y a <b>pas de texture en mmoire</b> ou 
   * que cette <b>mthode est non dfinie</b> par le matriel, la couleur retourne sera celle de la mthode <b>specularColor()</b>.
   * 
   * @param uv - La coordonne <i>uv</i> de la texture. 
   * @return La couleur de rflexion speculaire de la texture (si elle est disponible).
   */
  public SColor specularColor(SVectorUV uv);
  
	/**
	 * <p>Mthode pour obtenir la <b>couleur transmise</b> (filtre) par la surface du matriel. Elle permet de faire des calculs
	 * de filtrage de la lumire.</p> <p>Par exemple, une couleur de transparence (1,1,1) fait tout traverser, une couleur de 
	 * transparence (0.8, 0.8, 0.8) fait tout traverser avec attnuation et une couleur de transparence (1,0,0)
	 * fait uniquement traverser la couleur rouge.</p> Cette couleur est calcule  partir du produit de la couleur de la surface <b><i>S</i></b> avec la constant
	 * de transmission <b><i>k</i>t</b> tel que <ul><b><i>S</i>t = <i>k</i>t*<i>S</i></b>.</ul>
	 * 
	 * @return La couleur de la lumire <b><i>S</i>t</b> transmise par la surface.
	 */
	public SColor transparencyColor();
	
	/**
	 * Mthode pour obtenir le niveau de brillance de la surface du matriel.
	 * @return le niveau de brillance du matriel.
	 */
	public double getShininess();
	
	/**
	 * Mthode qui dtermine si le matriel rflchi la lumire comme un miroir.
	 * @return <b>true</b> si le matriel est rflchissant et <b>false</b> sinon.
	 */
	public boolean isReflective();
	
	/**
	 * Mthode pour obtenir le niveau de rflexion d'une source de lumire par la surface du matriel.
	 * @return le niveau de rflexion du matriel (0 = 0% rflchie, 1 = 100% rflchie).
	 */
	public double reflectivity();
	
	/**
	 * Mthode qui dtermine si le matriel est transparent.
	 * @return <b>true</b> si le matriel est transparent et <b>false</b> sinon.
	 */
	public boolean isTransparent();
	
	/**
	 * Mthode pour obtenir le niveau de transparence de la surface du matriel.
	 * @return le niveau de transparence du matriel (0 = 0% transparent, 1 = 100% transparent).
	 */
	public double transparency();
	
	/**
	 * Mthode pour obtenir le niveau d'opacit de la surface du matriel. Cette valeur correspond  l'inverse de la transparence.
	 * La somme de la transparence <b><i>k</i>t</b> avec l'opacit <b><i>k</i>o</b> donne 1 (kt + ko = 1).
	 * @return le niveau d'opacit du matriel (0 = 0% opaque, 1 = 100% opaque).
	 */
	public double opacity();
	
	/**
	 * Mthode pour obtenir l'indice de rfraction du matriel.
	 * @return L'indice de rfraction du matriel. Cependant, si le matriel est <b>opaque</b>, l'indice de rfraction sera automatiquement gale  <b>1.0</b>.
	 * @throws SRuntimeException Si le matriel n'est pas transparent.
	 */
	public double refractiveIndex();
	
	/**
	 * Mthode permettant de prciser si le mattiel contient des textures dans sa dfinition.
	 * 
	 * @return <b>true</b> si le matriel est dfini avec des textures et <b>false</b> sinon.
	 */
	public boolean asTexture();
	
}//fin interface SMaterial
