/**
 * 
 */
package sim.graphics.shader;

import java.util.List;

import sim.exception.SRuntimeException;
import sim.geometry.SRay;
import sim.geometry.space.SGeometrySpace;
import sim.graphics.SColor;
import sim.graphics.light.SLight;
import sim.graphics.material.SMaterial;

/**
 * La classe <b>SPhongReflexionShader</b> reprsente un shader selon le modle de rflexion de phong. 
 * Ce shader fait l'analyse d'un rayon principal avec un modle de rflexion ambiante, diffuse et spculaire.
 * 
 * @author Simon Vzina
 * @since 2015-02-03
 * @version 2016-05-06
 */
public class SPhongReflexionShader extends SAbstractShader {

	/**
	 * Constructeur d'un shader selon le modle de rflexion de phong avec le modle de rflexion spculaire de <b><i>Blinn</i></b>.
	 * @param geometry_space - L'espace des gomtries.
	 * @param t_max - Temps de dplacement maximal d'un rayon.
	 * @param light_list - La liste des sources de lumires.
	 */
	public SPhongReflexionShader(SGeometrySpace geometry_space, double t_max, List<SLight> light_list)
	{
		super(geometry_space, t_max, light_list);
	}

	/**
	 * Constructeur d'un shader selon le modle de rflexion de phong.
	 * @param geometry_space - L'espace des gomtries.
	 * @param t_max - Temps de dplacement maximal d'un rayon.
	 * @param light_list - La liste des sources de lumires.
	 * @param specular_reflexion_algo - Le type d'algorithme pour raliser le calcul de la rflexion spculaire.
	 */
	public SPhongReflexionShader(SGeometrySpace geometry_space, double t_max, List<SLight> light_list, int specular_reflexion_algo)
	{
		super(geometry_space, t_max, light_list, specular_reflexion_algo);
	}
	
	@Override
	public SColor shade(SRay ray) throws SRuntimeException
	{
		if(ray.asIntersected())
			throw new SRuntimeException("Erreur SPhongReflexionShader 001 : Le rayon a dj intersect une gomtrie pralablement.");
	
		// Effectuer l'intersection du rayon avec l'espace des gomtries
		ray = geometry_space.nearestIntersection(ray, t_max);
		
		// Vrification de l'intersection et retourner la couleur noire s'il n'y a pas d'intersection
		if(!ray.asIntersected())
			return SIllumination.NO_ILLUMINATION;
		
		 // Obtenir le matriel de la primitive
    SMaterial material = ray.getGeometry().getPrimitiveParent().getMaterial(); 
    
    // Vrifier s'il y a un algorithme d'illumination (pas en mode no_light)
    if(reflexion_algo == NO_LIGHT)
      if(ray.asUV())
        return material.diffuseColor(ray.getUV()); 
      else
        return material.diffuseColor();  
    
		// Retourner l'illumination directe
		return directIllumination(ray, material);
	}

	/**
	 * Mthode permettant d'valuer la couleur attribue  un rayon par illumination directe.
	 * 
	 * @param ray Le rayon ayant ralis une intersection avec une gomtrie de l'espace des gomtries.
	 * @param material Le matriel appliqu sur la gomtrie.
	 * @return La couleur de l'illumination directe.
	 */
	protected SColor directIllumination(SRay ray, SMaterial material)
	{
	  // Faire une somme de couleur et dbuter avec une couleur noire
    SColor color = SIllumination.NO_ILLUMINATION; 
                    
    // Iterer sur l'ensemble des lumires et ajouter leur contribution  la somme des couleurs.
    for(SLight light : light_list)
      color = color.add(shadeWithLight(ray, light, material));
      
    return color;
	}
	
}//fin classe SPhongReflexionShader
