/**
 * 
 */
package sim.loader.model;

import sim.exception.SConstructorException;
import sim.exception.SRuntimeException;
import sim.geometry.SConeGeometry;
import sim.geometry.SCylinderGeometry;
import sim.geometry.SGeometry;
import sim.geometry.STubeGeometry;
import sim.graphics.SColor;
import sim.graphics.SModel;
import sim.graphics.SPrimitive;
import sim.graphics.material.SBlinnMaterial;
import sim.graphics.material.SMaterial;
import sim.loader.SLoaderException;
import sim.loader.SStringLoader;
import sim.math.SVector3d;
import sim.parser.model.agp.SModelAGPParser;
import sim.parser.model.agp.SModelAGPParserException;
import sim.parser.model.agp.SPoint;
import sim.parser.model.agp.SPointSequence;
import sim.util.SLog;
import sim.util.SStringUtil;

/**
 * La classe <b>SModelAGPLoader</b> permet de charge un modle de format .agp tant le format de modle du <i>projet Anaglyphe</i> ralis par Anik Soulire.
 * 
 * @author Simon Vzina
 * @since 2015-08-01
 * @version 2015-05-12
 */
public class SModelAGPLoader implements SStringLoader {

  //--------------
  // CONSTANTES //
  //--------------
  
  /**
   * La constante <b>FILE_EXTENSION</b> correspond au nom de l'extension du fichier de lecture de ce type de modle 3d.
   */
  public static final String FILE_EXTENSION = "agp";  
  
  /**
   * La constante <b>DEFAULT_COLOR</b> correspond  la couleur par dfaut du modle 3d.
   */
  private static final SColor DEFAULT_COLOR = new SColor(0.8, 0.8, 0.2);
  
  /**
   * La constante <b>DEFAULT_RAY</b> correspond au rayon par dfaut des segment utilis pour reprsenter le modle 3d.
   */
  private static final double DEFAULT_RAY = 0.1;
  
  /**
   * La constante <b>TUBE_SEGMENT_CODE</b> correspond au code de rfrence pour construire des segments du modle 3d  l'aide de tube.
   */
  public static final int TUBE_SEGMENT_CODE = 0;
  
  /**
   * La constante <b>CYLINDER_SEGMENT_CODE</b> correspond au code de rfrence pour construire des segments du modle 3d  l'aide de cylindre.
   */
  public static final int CYLINDER_SEGMENT_CODE = 1;
 
  /**
   * La constante <b>CONE_SEGMENT_CODE</b> correspond au code de rfrence pour construire des segments du modle 3d  l'aide de cylindre.
   */
  public static final int CONE_SEGMENT_CODE = 3;
  
  /**
   * La constante <b>DEFAULT_SEGMENT_CODE</b> correspond au code de rfrence pour constuire des segments par dfaut tat gal  {@value}.
   */
  public static final int DEFAULT_SEGMENT_CODE = TUBE_SEGMENT_CODE;
  
  //-------------
  // VARIABLES //
  //-------------
  
  /**
   * La variable <b>segment_code</b> correspond au code de rfrence utilis pour choisir quel type 
   * de segement sera construit lors de la construction du modle 3d.
   */
  private final int segment_code;
  
  /**
   * La variable <b>error</b> correspond au nombre d'erreur lors de lecture du modle 3d.
   */
  private int error;
  
  //-----------------
  // CONSTRUCTUEUR //
  //-----------------
  
  /**
   *  Constructeur d'un chargeur de modle 3d de format agp avec segment de forme par dfaut. 
   */
  public SModelAGPLoader()
  {
    this(DEFAULT_SEGMENT_CODE);
  }

  /**
   * Constructeur d'un chargeur de modle 3d de format agp. 
   * 
   * @param segment_code Le code de rfrence pour dterminer le type de segment qui sera construit pour raliser le modle 3d.
   * @throws SConstructorException Si le code de rfrence n'est pas reconnu.
   */
  public SModelAGPLoader(int segment_code) throws SConstructorException
  {
    // Vrification du code de segment et affectation si valide
    switch(segment_code)
    {
      case TUBE_SEGMENT_CODE :
      case CYLINDER_SEGMENT_CODE : 
      case CONE_SEGMENT_CODE :      this.segment_code = segment_code;  break;
      
      default : throw new SConstructorException("Erreur SModelAGPLoader 001 : Le code de segment '" + segment_code + "' n'est pas reconnu.");
    }    
    
    error = 0;
  }
  
  //------------
  // MTHODES //
  //------------
  
  @Override
  public Object load(String string) throws SLoaderException
  {
    try{
      
      SModelAGPParser parser = new SModelAGPParser(string);
      
      SModel model = new SModel(string);
      SMaterial material = new SBlinnMaterial(DEFAULT_COLOR);
      
      // Construire des cylindres  partir de la liste de point des squences.
      // Puisque les points sont dans l'ordre, il y aura une chaine de cylindre.
      for(SPointSequence s : parser.getListSequence())
      {
        SPoint p1 = s.getPoint(0);  //le 1ier point de la squence
        SPoint p2;
        
        // Faire un segment avec comme point p2 le suivant, puis p1 devient p2 pour le prochain  faire 
        for(int i = 1; i < s.size(); i++)
        {
          p2 = s.getPoint(i);
             
          try{
           
            // Ajouter la primitive au modle
            model.addPrimitive(new SPrimitive(buildSegment(p1, p2, DEFAULT_RAY), material));
          
          }catch(SConstructorException e){
            // Compter le nombre d'erreur dans le modle (des points qui se sont rpts dans la squence) 
            error++;
          }
               
          p1 = p2;
        }
      }
      
      // Afficher un message dnombrant le nombre d'erreurs
      if(error > 0)
        SLog.logWriteLine("Message SModelAGPLoader : Le modle " + string + " contient " + error + " erreurs dans la dfinition du modle."); 
      
      return model;
      
    }catch(SModelAGPParserException e){
      throw new SLoaderException("Erreur SModelAGPLoader 002 : Une erreur lors de la lecture est survenue ce qui empche le chargement du modle 3d. " + SStringUtil.END_LINE_CARACTER + e.getMessage());
    }
  }
  
  /**
   * Mthode pour contruire le type de segment adquat pour reprsente le modle 3d.
   * 
   * @param p1 Le 1ier point du segment.
   * @param p2 Le 2ime point du segment.
   * @param ray Le rayon du segment.
   * @return La gomtrie correspondant au segment construit.
   * @throws SRuntimeException Si le code de rfrence du segment n'est pas trait.
   */
  private SGeometry buildSegment(SPoint p1, SPoint p2, double ray) throws SRuntimeException
  {
    switch(segment_code)
    {
      case TUBE_SEGMENT_CODE :      return new STubeGeometry(pointToVector(p1), pointToVector(p2), ray); 
        
      case CYLINDER_SEGMENT_CODE :  return new SCylinderGeometry(pointToVector(p1), pointToVector(p2), ray);
      
      case CONE_SEGMENT_CODE :      return new SConeGeometry(pointToVector(p1), pointToVector(p2), ray);
      
      default : throw new SRuntimeException("Erreur SModelAGPLoader 003 : Le code de segment '" + segment_code + "' n'est pas reconnu.");
    }  
  }
  
  /**
   * Mthode pour convertir un SPoint en SVector3d.
   * 
   * @param p Le point  convertir.
   * @return Le vecteur associ au point.
   */
  private SVector3d pointToVector(SPoint p)
  {
    return new SVector3d(p.getX(), p.getY(), p.getZ());
  }
  
}//fin de la classe SModelAGPLoader
