/**
 * 
 */
package sim.parser.model.agp;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.StringTokenizer;

import sim.parser.model.obj.SModelOBJParserException;
import sim.util.SBufferedReader;
import sim.util.SFileSearch;
import sim.util.SStringUtil;


/**
 * Classe reprsentant un lecteur de fichier .agp correspondant  un modle 3d dans le format du <i>projet Anaglyphe</i> ralis par Anik Soulire.
 * 
 * @author Simon Vzina
 * @since 2015-08-01
 * @version 2015-08-02
 */
public class SModelAGPParser {

  private static final String MODEL_EXTENSION = "agp";
  
  private final String file_name;                     //nom du fichier .agp
  private final List<SPointSequence> sequence_list;   //liste des squences de points
  
  /**
   * Constructeur d'un parser pour objet 3d de de format agp.
   * @param file_name - Le nom du fichier comportant le modle 3d de format obj.
   * @throws SModelOBJParserException Si le fichier n'est pas adquat. 
   */
  public SModelAGPParser(String file_name)throws SModelAGPParserException
  {
    this.file_name = file_name;
    sequence_list = new ArrayList<SPointSequence>();
    
    parseAGP(file_name);
  }
  
  /**
   * Mthode pour obtenir le nom du fichier du modle 3d de format AGP.
   * @return Le nom du fichier.
   */
  public String getFileName()
  {
    return file_name;
  }
  
  /**
   * Mthode pour obtenir la liste des squences de points du modle 3d de format AGP.
   * @return La liste des squences.
   */
  public List<SPointSequence> getListSequence()
  {
    return sequence_list;
  }
  
  /**
   * Mhode pour faire la lecture d'un fichier agp  partir du nom du fichier.
   * @param file_name - Le nom du fichier agp  lire.
   * @throws SModelAGPParserException Si le fichier n'est pas dans le bon format.
   * @throws SModelAGPParserException Si le fichier n'est pas trouv.
   * @throws SModelAGPParserException Si le fichier existe sous plusieurs versions dans les diffrents sous-rpertoires.
   */
  private void parseAGP(String file_name)throws SModelAGPParserException
  {
    //Vrification du format du fichier
    if(!SStringUtil.extensionFileLowerCase(file_name).equals(MODEL_EXTENSION))
      throw new SModelAGPParserException("Erreur SModelAGPParser 001 : Le fichier '" + file_name + "' n'est pas dans le format '" + MODEL_EXTENSION + "'.");
    
    //Trouver le fichier  partir du rpertoire o l'excution de l'application est ralise
    SFileSearch search = new SFileSearch("", file_name);
    
    if(!search.isFileFound())
      throw new SModelAGPParserException("Erreur SModelAGPParser 002 : Le fichier '" + file_name + "' n'est pas trouv.");
    
    if(search.isManyFileFound())
      throw new SModelAGPParserException("Erreur SModelAGPParser 003 : Le fichier '" + file_name + "' a t trouv plus d'une fois dans les diffrents sous-rpertoires. Veuillez en garder qu'une seule version.");
    
    File file = new File(search.getFileFoundList().get(0)); //prendre la 1ier trouv dans la liste
    
    try
    {
      FileReader fr = new FileReader(file);
      SBufferedReader sbr = new SBufferedReader(fr);
      parseAGP(sbr);
      
    }catch(FileNotFoundException e){
      throw new SModelAGPParserException("Erreur SModelAGPParser 004 : IMPOSSIBLE! La localisation du fichier a t dj test.");     
    }catch(IOException ioe){
      throw new SModelAGPParserException("Erreur SModelAGPParser 005 : Une erreur de type I/O est survenue.");      
    }
  }
  
  /**
   * Mthode pour lire le fichier de format agp  l'aide d'un SBufferedReader.
   * 
   * @param sbr - Le buffer de lecture.
   * @throws IOException S'il y a eu une erreur de type I/O lors de la lecture du fichier.
   * @throws SModelAGPParserException S'il y a eu une erreur menant  une incapacit de dfinir le modle 3d.
   */
  private void parseAGP(SBufferedReader sbr)throws SModelAGPParserException, IOException
  { 
    boolean end_reading = false;        //condition de fin de lecture
    
    while(!end_reading)
    {
      String line = sbr.readLine();    //ligne courante en lecture
      
      //Si la ligne lu est "null", la fin du fichier est atteinte
      if(line == null)
        end_reading = true;
      else
      {
        StringTokenizer tokens = new StringTokenizer(line);  //fragmenter la ligne lue en liminant les caractres d'espacement
        
        //S'il y a un token  analyser
        if(tokens.hasMoreTokens())
        {
          String first_token = tokens.nextToken();        //obtenir le 1ier token  analyser
          first_token = first_token.replaceAll(",",".");  //remplacer les ',' par des '.' car c'est le format employ par java
          
          //Si le 1ier token peut tre converti en double, c'est que l'on commence la dfinition d'une squence.
          //Si la conversion n'est pas possible, c'est que nous avons une exception correspondant  une ligne de commentaire.
          //Nous allons alors analyser :
          //1ire ligne - Coordonne x de la squence de points
          //2ime ligne - Coordonne y de la squence de points
          //3ime ligne - Coordonne z de la squence de points
          try{
            
            Double.parseDouble(first_token);  //vrifier si le 1ier token est un double, une exception sera lance si ce n'est pas le cas
             
            String line_x = line;                 //ligne dj lue
            line_x = line_x.replaceAll(",",".");  //remplacement des ',' par des '.' pour les nombres dcimaux
                        
            String line_y = sbr.readLine();       //ligne suivante lue
            line_y = line_y.replaceAll(",",".");  //remplacement des ',' par des '.' pour les nombres dcimaux
                                    
            String line_z = sbr.readLine();       //ligne suivante lue
            line_z = line_z.replaceAll(",",".");  //remplacement des ',' par des '.' pour les nombres dcimaux
                        
            parsePointSequence(line_x, line_y, line_z, sbr.atLine() - 3);
            
          }catch(NumberFormatException e){
            //rien faire, car c'est une ligne de commentaire ...
          }
        }
      }  
    }//fin while
  }
  
  /**
   * Mthode pour analyser trois string dfinissant les coordonnes x,y et z d'une squence de points.
   * @param line_x - La ligne dfinissant les coordonnes x.
   * @param line_y - La ligne dfinissant les coordonnes y.
   * @param line_z - La ligne dfinissant les coordonnes z.
   * @throws SModelAGPParserException Si les trois string n'ont pas le mme nombre d'lment en dfinition.
   * @throws SModelAGPParserException Si une ligne n'est pas bien dfini.
   */
  public void parsePointSequence(String line_x, String line_y, String line_z, long reading_line)throws SModelAGPParserException
  {
    //Vrification si une ligne est 'null'. Dans ce cas, il y a une dernire squence de point qui est alors mal dfini
    if(line_x == null || line_y == null || line_z == null)
      throw new SModelAGPParserException("Erreur SModelAGPParseur 006 :  la squence de ligne " + reading_line + ", une ligne est manquante ce qui rend la squence de points mal dfinie.");
    
    //Fragmenter les lignes
    StringTokenizer tokens_x = new StringTokenizer(line_x);  //fragmenter la ligne dfinissant x
    StringTokenizer tokens_y = new StringTokenizer(line_y);  //fragmenter la ligne dfinissant y
    StringTokenizer tokens_z = new StringTokenizer(line_z);  //fragmenter la ligne dfinissant z
    
    //Vrification sur le nombre d'lment par ligne
    if(tokens_x.countTokens() != tokens_y.countTokens() || tokens_x.countTokens() != tokens_z.countTokens())
      throw new SModelAGPParserException("Erreur SModelAGPParseur 007 :  la squence de ligne " + reading_line + ", il y a " +  tokens_x.countTokens() + " coordonnes x, " + tokens_y.countTokens() + " coordonnes y et " + tokens_z.countTokens() + " coordonnes z. La squence est mal dfinie.");
  
    try{
      
      //Construire la squence de point avec le nombre de point correspondant au nombre de tokens
      SPointSequence sequence = new SPointSequence(tokens_x.countTokens());
      
      int index;  //l'indexage des points
      
      //Remplir la squence de coordonne x
      index = 0;
      
      while(tokens_x.hasMoreTokens())
      {
        sequence.setX(index, Double.parseDouble(tokens_x.nextToken()));
        index++;
      }
      
      //Remplir la squence de coordonne y
      index = 0;
      
      while(tokens_y.hasMoreTokens())
      {
        sequence.setY(index, Double.parseDouble(tokens_y.nextToken()));
        index++;
      }
      
      //Remplir la squence de coordonne z
      index = 0;
      
      while(tokens_z.hasMoreTokens())
      {
        sequence.setZ(index, Double.parseDouble(tokens_z.nextToken()));
        index++;
      }
      
      //La squence tant termine, nous pouvons l'ajouter  la liste des squences
      sequence_list.add(sequence);
      
    }catch(NumberFormatException e){
      throw new SModelAGPParserException("Erreur SModelAGPParseur 008 :  la squence de ligne " + reading_line + ", un lment lu n'a pas t convertir en valeur numrique. La squence est ainsi invalide.");
    }
  }
  
  /**
   * Test de la classe SModelAGPParser
   * @param arg
   */
  public static void main(String[] arg)
  {
    test("shuriken.agp");
    test("pyramide.agp");
  }
  
  private static void test(String file_name)
  {
    try{
      
      System.out.println("Fichier en test : '" + file_name + "'.");
      
      SModelAGPParser parser = new SModelAGPParser(file_name);
    
      //Faire l'affichage du contenu
      List<SPointSequence> sequences = parser.getListSequence();
      
      for(SPointSequence s : sequences)
      { 
        for(int i=0; i<s.size(); i++)
          System.out.print(s.getPoint(i).getX() + "\t");
        System.out.println();
        
        for(int i=0; i<s.size(); i++)
          System.out.print(s.getPoint(i).getY() + "\t");
        System.out.println();
        
        for(int i=0; i<s.size(); i++)
          System.out.print(s.getPoint(i).getZ() + "\t");
        System.out.println();
        
        System.out.println();
      }
    }catch(SModelAGPParserException e){
      e.printStackTrace();
      System.out.println("test en chec !!!");
    }
  }
}//fin de la classe SModelAGPParser
