/**
 * 
 */
package sim.parser.model.obj;

/**
 * Classe qui reprsente la description d'un matriel pour un objet 3d de format OBJ (Wavefront).
 * @author Simon Vzina
 * @since 2015-03-28
 * @version 2015-11-05
 */
public class SMaterialOBJ {

  //Paramtres par dfaut accessibles publiquement
  public static final String DEFAULT_NAME = "default";
  public static final SVertex DEFAULT_AMBIENT_COEFFICIENT = new SVertex(0.2f, 0.2f, 0.2f);  
  public static final SVertex DEFAULT_DIFFUSE_COEFFICIENT = new SVertex(0.3f, 0.7f, 0.3f);
  public static final SVertex DEFAULT_SPECULAR_COEFFICIENT = new SVertex(0.2f, 0.2f, 0.2f);
  public static final float DEFAULT_SHININESS = 5.0f;
  
  private final String material_name; //nom du matriel
  
  private final SVertex Ka; //coefficient de rflexion ambiante
  private final SVertex Kd; //coefficient de rflexion diffuse
  private final SVertex Ks; //coefficient de rflexion spculaire
  private final float Ns;   //constante de brillance (shininess)
  
  private final String texture_Ka;  //nom de la texture de couleur ambiante
  private final String texture_Kd;  //nom de la texture de couleur diffuse
  private final String texture_Ks;  //nom de la texture de couleur spculaire
  
  /**
   *  Constructeur d'un matriel par dfaut pour objet 3d de format OBJ.
   */
  public SMaterialOBJ()
  {
    this(DEFAULT_NAME, DEFAULT_AMBIENT_COEFFICIENT, DEFAULT_DIFFUSE_COEFFICIENT, DEFAULT_SPECULAR_COEFFICIENT, DEFAULT_SHININESS);
  }
  
  /**
   * Constructeur d'un matriel pour objet 3d de format OBJ.
   * 
   * @param material_name - Le nom du matriel.
   * @param Ka - Le coefficient de rflexion ambiante pour les composantes rgb.
   * @param Kd - Le coefficient de rflexion diffuse pour les composantes rgb.
   * @param Ks - Le coefficient de rflexion spculaire pour les composantes rgb.
   * @param Ns - La constance de brillance (shininess).
   */
  public SMaterialOBJ(String material_name, SVertex Ka, SVertex Kd, SVertex Ks, float Ns)
  {
    this(material_name, Ka, Kd, Ks, Ns, null, null, null);
  }
  
  /**
   * Constructeur d'un matriel pour objet 3d de format OBJ avec texture.
   * 
   * @param material_name - Le nom du matriel.
   * @param Ka - Le coefficient de rflexion ambiante pour les composantes rgb.
   * @param Kd - Le coefficient de rflexion diffuse pour les composantes rgb.
   * @param Ks - Le coefficient de rflexion spculaire pour les composantes rgb.
   * @param Ns - La constance de brillance (shininess).
   * @param texture_Ka - Le nom du fichier comprenant l'information de la texture pour la rflexion ambiante.
   * @param texture_Kd - Le nom du fichier comprenant l'information de la texture pour la rflexion diffuse.
   * @param texture_Ks - Le nom du fichier comprenant l'information de la texture pour la rflexion spculaire.
   */
  public SMaterialOBJ(String material_name, SVertex Ka, SVertex Kd, SVertex Ks, float Ns, String texture_Ka, String texture_Kd, String texture_Ks)
  {
    this.material_name = material_name;
    
    this.Ka = Ka;
    this.Kd = Kd;
    this.Ks = Ks;
    this.Ns = Ns;
    
    this.texture_Ka = texture_Ka;
    this.texture_Kd = texture_Kd;
    this.texture_Ks = texture_Ks;
  }
  
  /**
   * Mthode pour obtenir le nom du matriel. Le nom par dfaut sera "default" si le matriel
   * a t construit sans nom particulier.
   * @return Le nom du matriel.
   */
  public String getMaterialName()
  {
    return material_name;
  }
  
  /**
   * Mthode pour obtenir les coefficients de rflexion ambiante du matriel.
   * @return Le vertex des coefficients de rflexion ambiante.
   */
  public SVertex getKa()
  {
    return Ka;
  }
  
  /**
   * Mthode pour obtenir les coefficients de rflexion diffuse du matriel.
   * @return Le vertex des coefficients de rflexion diffuse.
   */
  public SVertex getKd()
  {
    return Kd;
  }
  
  /**
   * Mthode pour obtenir les coefficients de rflexion spculaire du matriel.
   * @return Le vertex des coefficients de rflexion spculaire.
   */
  public SVertex getKs()
  {
    return Ks;
  }
  
  /**
   * Mthode pour obtenir le coefficient de brillance (shininess) du matriel.
   * @return Le coefficient de brillance (shininess).
   */
  public float getNs()
  {
    return Ns;
  }
  
  /**
   * Mthode pour obtenir le nom du fichier de la texture pour la rflexion ambiante.
   * @return Le nom du fichier comprenant l'information de la texture pour la rflexion ambiante. 
   * Le nom du fichier sera <b>null</b> s'il n'y a pas de texture affecte.
   */
  public String getTextureKaFileName()
  {
    return texture_Ka;
  }
  
  /**
   * Mthode pour obtenir le nom du fichier de la texture pour la rflexion diffuse.
   * @return Le nom du fichier comprenant l'information de la texture pour la rflexion diffuse.
   * Le nom du fichier sera <b>null</b> s'il n'y a pas de texture affecte.
   */
  public String getTextureKdFileName()
  {
    return texture_Kd;
  }
  
  /**
   * Mthode pour obtenir le nom du fichier de la texture pour la rflexion spculaire.
   * @return Le nom du fichier comprenant l'information de la texture pour la rflexion spculaire.
   * Le nom du fichier sera <b>null</b> s'il n'y a pas de texture affecte.
   */
  public String getTextureKsFileName()
  {
    return texture_Ks;
  }
  
  /**
   * Mthode pour dterminer si le matriel contient une rfrence  une texture. 
   * @return <b>true</b> si la matriel contient une texture et <b>false</b> sinon.
   */
  public boolean asTexture()
  {
    if(texture_Ka != null)
      return true;
    
    if(texture_Kd != null)
      return true;
    
    if(texture_Ks != null)
      return true;
    
    return false;
  }
  
}//fin SMaterialOBJ
