/**
 * 
 */
package sim.physics;

import sim.exception.SConstructorException;
import sim.math.SVector3d;

/**
 * La classe <b>SOsillator</b> reprsente un oscillateur harmonique simple. 
 * 
 * @author Simon Vzina
 * @since 2016-02-14
 * @version 2016-02-23
 */
public class SOscillator {
  
  //--------------
  // CONSTANTES //
  //--------------
  
  /**
   * La constante <b>DEFAULT_AMPLITUDE</b> correspond  l'amplitude de l'oscillateur par dfaut.
   */
  private static final double DEFAULT_AMPLITUDE = 1.0;
  
  /**
   * La constante <b>DEFAULT_PHI</b> correspond  la constante de phase par dfaut.
   */
  private static final double DEFAULT_PHI = 0.0;
  
  //-------------
  // VARIABLES //
  //-------------
  
  /**
   * La variable <b>position</b> correspond  la position de l'oscillateur
   */
  private final SVector3d position;

  /**
   * La variable <b>frequency</b> correspond  la frquence de l'oscillateur. 
   */
  private final double frequency;
  
  /**
   * La variable <b>A</b> correspond  l'amplitude de l'oscillateur.
   */
  private final double A;
  
  /**
   * La variable <b>phi</b> correspond  la constante de phase des ondes gnres par l'oscillateur.
   */
  private final double phi;
  
  //-------------
  // Prcalcul //
  //-------------
  
  /**
   * La variable <b>omega</b> correspond  la frquence angulaire des ondes gnres par l'oscillateur.
   */
  private final double omega;
  
  //----------------
  // CONSTRUCTEUR //
  //----------------
  
  /**
   * Constructeur d'un oscillateur harmonique simple d'amplitude et de phase par dfaut. 
   * 
   * @param position La position de l'oscillateur.
   * @param frequency La frquence de l'oscillateur.
   */
  public SOscillator(SVector3d position, double frequency)
  {
    this(position, frequency, DEFAULT_AMPLITUDE, DEFAULT_PHI);
  }
  
  /**
   * Constructeur d'un oscillateur harmonique simple. 
   * 
   * @param position La position de l'oscillateur.
   * @param frequency La frquence de l'oscillateur.
   * @param amplitude L'amplitude de l'oscillateur.
   * @param phase La phase initiale de l'oscillateur.
   * @throws SConstructorException Si la frquence des oscillations est ngatives.
   */
  public SOscillator(SVector3d position, double frequency, double amplitude, double phase) throws SConstructorException
  {
    // Vrification de la longueur d'onde de l'oscillateur
    if(frequency < 0)
      throw new SConstructorException("Erreur SOscillator 001 : La frquence '" + frequency + "' ne peut tre ngative.");
    
    this.position = position;
    
    this.frequency = frequency;
    
    A = amplitude;
    
    phi = SWaveOptics.phaseBetweenZeroAnd2Pi(phase);
    
    // valuer la frquence angulaire
    omega = 2*Math.PI*frequency;
  }

  //------------
  // MTHODES //
  //------------
  
  /**
   * Mthode pour obtenir l'amplitude de l'oscillateur harmonique simple.
   * 
   * @return L'amplitude de l'oscillateur.
   */
  public double getAmplitude()
  {
    return A;
  }
  
  /**
   * Mthode pour obtenir la position de l'oscillateur.
   * 
   * @return La position de l'oscillateur.
   */
  public SVector3d getPosition()
  {
    return position;
  }
  
  /**
   * Mthode pour obtenir la frquence des oscillations de l'oscillateur.
   * 
   * @return La frquence de l'oscillateur.
   */
  public double getFrequency()
  {
    return frequency;
  }
  
  /**
   * Mthode pour obtenir la priodes des oscillations de l'oscillateur.
   * 
   * @return La priode (en seconde).
   */
  public double getPeriod()
  {
    return 1.0 / frequency;
  }
   
  /**
   * Mthode pour obtenir la valeur de l'onde  l'endroit o le point P a t dsign (par la mthode <i>setEvaluatePosition(SVector3d p)</i>).
   * 
   * @param t Le temps.
   * @return La valeur de l'oscillateur au temps t.
   */
  public double getValue(double t)
  {
    // Pour des raisons numriques, je vais valuer omega*t et le "recadrer" entre 0 et 2Pi
    return A*Math.sin( SWaveOptics.phaseBetweenZeroAnd2Pi(omega*t) + phi);
  }
  
  /**
   * Mthode pour obtenir la phase initiale (constante de phase) de l'oscillateur. 
   * Ceci correspond au paramtre <b>phi</b> de l'oscillateur dans l'expression
   * <ul>x = A sin (omega*t + phi)</ul>
   * 
   * @return La constante de phase de l'oscillateur entre zro et 2*Pi.
   */
  public double getInitialPhase()
  {
    return phi;
  }
  
  /**
   * Mthode pour obtenir la phase temporelle de l'oscillateur.
   * Ceci correspond au paramtre <b>omega*t</b> de l'oscillateur dans l'expression
   * <ul>x = A sin (omega*t + phi)</ul>
   * 
   * @param t Le temps de l'oscillation de l'oscillateur.
   * @return La phase temporelle de l'oscillateur entre zro et 2*Pi.
   */
  public double getTimePhase(double t)
  {
    return SWaveOptics.phaseBetweenZeroAnd2Pi(omega*t);
  }
  
}//fin de la classe SOscillator
